"""
Timber
------

This module provides useful functions to conveniently wrap the functionality of ``pytimber``.
"""
import logging

from omc3.utils.mock import cern_network_import
from omc3.utils.time_tools import AccDatetime

pytimber = cern_network_import("pytimber")

LOG = logging.getLogger(__name__)


def find_exact_time_for_beamprocess(acc_time: AccDatetime) -> AccDatetime:
    """
    Finds the last entry where HX:SRMP-POW equals 123. Which is, according to the online model
    KnobExtractor, the timing event for SQUEEZE or RAMP.
    I don't think this is correct (jdilly).
    """
    db = pytimber.LoggingDB(source="nxcals")
    t1, t2 = acc_time.sub(days=1).local_string(), acc_time.local_string()

    event_ts, event_val = db.get("HX:SRMP-POW", t1, t2)["HX:SRMP-POW"]
    event_ts = event_ts[event_val == 123.0]

    if len(event_ts) == 0:
        raise ValueError(
            f"No valid beamprocess found in the 24h before {acc_time.cern_utc_string()}"
        )

    exact_time = acc_time.__class__.from_timestamp(event_ts[-1])
    LOG.debug(f"Exact time for beamprocess found: {exact_time.cern_utc_string()}")
    return exact_time



"""
Additional IO-Tools
---------------------

Additional tools for reading and writing TfsDataFrames,
that are not necessarily related to TFS-files.
"""
import pandas as pd
from pathlib import Path

from typing import Union

import h5py
from tfs import TfsDataFrame

import logging

LOGGER = logging.getLogger(__name__)


def write_hdf(path: Union[Path, str], df: TfsDataFrame, **kwargs):
    """Write TfsDataFrame to hdf5 file. The dataframe will be written into
    the group ``data``, the headers into the group ``headers``.
    Only one frame per file is allowed.

    Args:
        path (Path, str): Path of the output file.
        df (TfsDataFrame): TfsDataFrame to write.
        kwargs: kwargs to be passed to pandas ``DataFrame.to_hdf()``.
                ``key`` is not allowed and ``mode`` needs to be ``w`` if the
                output file already exists.
    """
    # Check for `key` kwarg (forbidden) ---
    if "key" in kwargs:
        raise AttributeError("The argument 'key' is not allowed here, "
                             "as only one TfsDataFrame per file is supported.")

    # Check for `mode` kwarg (allowed under circumstances but generally ignored) ---
    user_mode = kwargs.pop('mode', None)
    if user_mode is not None and user_mode != "w":
        if path.exists():
            raise AttributeError(f"'mode=\"{user_mode}\"' is not allowed here. "
                                 f"The output file at {str(path)} will always be overwritten!")
        LOGGER.warning(f"'mode=\"{user_mode}\"' is not allowed here. "
                       f"Mode \"w\" will be used.")

    # Actual writing of the output file ---
    df.to_hdf(path, key='data', mode='w', **kwargs)
    with h5py.File(path, mode="a") as hf:
        hf.create_group("headers")  # empty group in case of empty headers
        for key, value in df.headers.items():
            hf.create_dataset(f"headers/{key}", data=value)


def read_hdf(path: Union[Path, str]) -> TfsDataFrame:
    """Read TfsDataFrame from hdf5 file. The DataFrame needs to be stored
    in a group named ``data``, while the headers are stored in ``headers``.

    Args:
        path (Path, str): Path of the file to read.
    """
    df = pd.read_hdf(path, key="data")
    with h5py.File(path, mode="r") as hf:
        headers = hf.get('headers')
        headers = {k: headers[k][()] for k in headers.keys()}

    for key, value in headers.items():
        try:
            headers[key] = value.decode('utf-8')  # converts byte-strings back
        except AttributeError:
            pass  # probably numeric
    return TfsDataFrame(df, headers=headers)



"""
PyLHC
~~~~~~~~~~~~~~~~

PyLHC is a script collection for the optics measurements and corrections group (OMC) at CERN.

:copyright: pyLHC/OMC-Team working group.
:license: MIT, see the LICENSE.md file for details.
"""
__title__ = "pylhc"
__description__ = "An accelerator physics script collection for the OMC team at CERN."
__url__ = "https://github.com/pylhc/pylhc"
__version__ = "0.6.0"
__author__ = "pylhc"
__author_email__ = "pylhc@github.com"
__license__ = "MIT"

__all__ = [__version__]



"""
Dispersion
----------

The functions in this script compute the calibration factors for the LHC BPMs
using the dispersion method. The `get_calibration_factors_from_dispersion` is
intended to be used with the script `bpm_calibration.py`.

"""
from pathlib import Path
from scipy.optimize import curve_fit
import numpy as np
import pandas as pd
from typing import Tuple, List, Dict

from omc3.utils import logging_tools
from omc3.optics_measurements.constants import (
    ERR,
    EXT,
    NORM_DISP_NAME,
    DISPERSION_NAME,
    S,
)

from pylhc.constants.calibration import (
    BPMS,
    D_BPMS,
    D,
    IPS,
    LABELS,
    ND,
    TFS_INDEX,
)
from pylhc.constants.general import PLANES
import tfs


LOG = logging_tools.get_logger(__name__)


def _get_dispersion_fit(
    positions: pd.Series, dispersion_values: pd.Series, dispersion_err: pd.Series
) -> Tuple[pd.Series, pd.Series]:
    """
    This function returns a fit of the given dispersion values along with the
    associated error.

    Args:
      positions (pd.Series): Positions of the BPMs to be fitted.
      dispersion_values (pd.Series): Values of the BPMs to be fitted.
      dispersion_err (pd.Series): Associated errors to the values.

    Returns:
      Tuple[pd.Series, pd.Series]: The elements returned are the values of the
      fit of the dispersion values and the associated error.
    """

    def dispersion_function(x, a, b):
        return a * x + b

    # Filter the values we have to only keep the asked BPMs
    values = dispersion_values[dispersion_values.index.isin(positions.index)]
    err = dispersion_err[dispersion_values.index.isin(positions.index)]

    # Get the curve fit for the expected affine function
    valid = ~(np.isnan(positions) | np.isnan(values))
    fit, fit_cov = curve_fit(dispersion_function, positions[valid], values[valid], sigma=err[valid])

    # Get the error from the covariance matrix
    fit_err = np.sqrt(np.diag(fit_cov))

    # Get the fitted beta and add the errors to get min/max values
    dispersion_fit = dispersion_function(positions, fit[0], fit[1])
    dispersion_max_fit = dispersion_function(
        positions, fit[0] + fit_err[0], fit[1] + fit_err[1]
    )
    dispersion_min_fit = dispersion_function(
        positions, fit[0] - fit_err[0], fit[1] - fit_err[1]
    )
    dispersion_fit_err = (dispersion_max_fit - dispersion_min_fit) / 2

    return dispersion_fit, dispersion_fit_err


def _get_factors_from_dispersion(
        dispersion: Dict[str, pd.Series], phase: str, phase_err: str,
) -> Tuple[pd.Series, pd.Series]:
    """
    This function computes the calibration factors for the dispersion method
    with the non fitted dispersion values. The associated error is also
    calculated.
    The equations being the same for the factors from dispersion and dispersion
    fit, this function can be used for both.

    Args:
      dispersion (Dict[str, pd.Series]): Dictionnary containing 4 keys: phase,
        phase_err, amp and amp_err. Each key is related to the method used to
        obtain the dispersion and its error. The keys differ for the dispersion
        fit method, as a _fit part is added.
      phase (str): key for the dispersion from phase
      phase_err (str): key for the dispersion from phase error

    Returns:
      Tuple[pd.Series, pd.Series]: The first Series are the calibration
        factors, the second one their error.
    """
    # Get the ratios, those are our calibration factors
    factors = dispersion[phase] / dispersion["amp"]

    # Code in BBs
    calibration_error = (dispersion[phase_err] / dispersion["amp"]) ** 2
    calibration_error += (
        dispersion["amp_err"] * (dispersion[phase] / (dispersion["amp"] ** 2))
    ) ** 2
    calibration_error = np.sqrt(calibration_error)

    return factors, calibration_error


def get_calibration_factors_from_dispersion(
    ips: List[int], input_path: Path
) -> Dict[str, pd.DataFrame]:
    """
    This function is the main function to compute the calibration factors for
    the dispersion method.
    Given an IP and a path containing the corresponding Tfs files, this
    function returns the calibration factors using both the dispersion and its
    fitted values.
    The calibration factors based on the dispersion are only computed for the
    X plane.

    Args:
      ips (List[int]): IPs to compute the calibration factors for.
      input_path (Path): Path of the directory containing the beta files.

    Returns:
      Dict[str, pd.DataFrame]: The returned DataFrame object contains the
      calibration factors for each BPM along with their error. Both the
      dispersion and dispersion from fit values are used, resulting in 6
      columns:

        - NAME: BPM Name
        - S: Position
        - CALIBRATION: Calibration factors computed from the dispersion
        - ERROR_CALIBRATION: Associated error to the above calibration
          factors
        - CALIBRATION_FIT: Calibration factors computed from the fitted
          dispersion
        - ERROR_CALIBRATION_FIT: Associated error to the above calibration
          factors
    """
    LOG.info("Computing the calibration factors via dispersion")
    # Load the normalized dispersion tfs file
    norm_dispersion_tfs = tfs.read(
        input_path / f"{NORM_DISP_NAME}x{EXT}", index=TFS_INDEX
    )
    dispersion_tfs = tfs.read(input_path / f"{DISPERSION_NAME}x{EXT}", index=TFS_INDEX)

    # Get the beam concerned by those tfs files
    beam = int(dispersion_tfs.iloc[0].name[-1])

    # Loop over the IPs and compute the calibration factors
    calibration_factors = dict()
    for ip in ips:
        LOG.info(f"  Computing the calibration factors for IP {ip}, plane X")
        # Filter our TFS files to only keep the BPMs for the selected IR
        bpms = dispersion_tfs.reindex(BPMS[ip][beam])
        d_bpms = dispersion_tfs.reindex(D_BPMS[ip][beam])
            
        # Check for possible missing bpms
        for bpm_set in [bpms, d_bpms]:
            missing = set(bpm_set.loc[bpm_set.isnull().values].index)
            if missing:
                LOG.warning("    One or several BPMs are missing in the input"
                            " DataFrame, the calibration factors calculation"
                            f"from fit may not be accurate: {missing}")

        # Get the positions of the BPMs and the subset used for the fit
        bpms = bpms.index
        d_bpms = d_bpms.index
        positions = dispersion_tfs.reindex(bpms)[S]
        positions_fit = dispersion_tfs.reindex(d_bpms)[S]

        # Get the dispersion and dispersion from phase from the tfs files
        dispersion = dict()
        normalised_dispersion = dict()

        dispersion["amp"] = dispersion_tfs.reindex(bpms)[f"DX"]
        dispersion["amp_err"] = dispersion_tfs.reindex(bpms)[f"{ERR}{D}X"]

        dispersion["phase"] = norm_dispersion_tfs.reindex(bpms)[f"DX"]
        dispersion["phase_err"] = norm_dispersion_tfs.reindex(bpms)[f"{ERR}{D}X"]

        # Compute the calibration factors using the dispersion from phase and amp
        calibration, calibration_err = _get_factors_from_dispersion(
            dispersion, "phase", "phase_err")

        # Fit the dispersion from phase
        dispersion["phase_fit"], dispersion["phase_fit_err"] = _get_dispersion_fit(
            positions_fit, dispersion["phase"], dispersion["phase_err"]
        )

        # Compute the calibration factors using the fitted dispersion from amp / phase
        calibration_fit, calibration_fit_err = _get_factors_from_dispersion(
            dispersion, "phase_fit", "phase_fit_err"
        )

        # Assemble the calibration factors in one dataframe
        factors_for_ip = pd.concat(
            [
                positions,
                calibration,
                calibration_err,
                calibration_fit,
                calibration_fit_err,
            ],
            axis=1,
        )
        factors_for_ip.columns = LABELS
        factors_for_ip.index.name = TFS_INDEX

        if "X" not in calibration_factors.keys():
            calibration_factors = {"X": factors_for_ip}
        else:
            calibration_factors["X"] = calibration_factors["X"].append(factors_for_ip)

    return calibration_factors



"""
Constants: KickGroups
---------------------

Constants used in the KickGroups
"""
from pathlib import Path

KICKGROUPS_ROOT = Path("/user/slops/data/LHC_DATA/OP_DATA/Betabeat/KickGroups/MULTITURN_ACQ_GROUPS")

# Columns ---
KICKGROUP = "KICKGROUP"
SDDS = "SDDS"
TURNS = "TURNS"
BUNCH = "BUNCH"
UTCTIME = "UTC"
TIMESTAMP = "TIMESTAMP"
LOCALTIME = "LOCAL"
TUNEX = "QX"
TUNEY = "QY"
DRIVEN_TUNEX = "DQX"
DRIVEN_TUNEY = "DQY"
DRIVEN_TUNEZ = "DQZ"
AMPX = "AMPX"
AMPY = "AMPY"
AMPZ = "AMPZ"
OPTICS = "OPTICS"
OPTICS_URI = "OPTICS_URI"
BEAMPROCESS = "BEAMPROCESS"
BEAM = "BEAM"

KICK_COLUMNS = [UTCTIME, LOCALTIME, TUNEX, TUNEY, DRIVEN_TUNEX, DRIVEN_TUNEY, DRIVEN_TUNEZ, AMPX, AMPY, AMPZ, TURNS, BUNCH, SDDS, BEAM, OPTICS, OPTICS_URI, BEAMPROCESS]
COLUMNS_TO_HEADERS = [BEAM, BUNCH, TURNS, BEAMPROCESS, OPTICS, OPTICS_URI]
KICK_GROUP_COLUMNS = [UTCTIME, LOCALTIME, KICKGROUP, TIMESTAMP]


"""
BSRT logger
-------------------

Script used during Run II to log detailed BSRT data and save it for later analysis. Data from the
BSRT for each timestep is put in a `dictionary` and append to a `list`. The `list` is then saved
to disk through pickling. Proper testing requires communication with ``FESA`` s class, possible
only from the Technical Network.

Original authors: E. H. Maclean, T. Persson and G. Trad.
"""
import datetime as dt
import os
import pickle
import sys
import time

from omc3.definitions import formats
from omc3.utils.mock import cern_network_import

pyjapc = cern_network_import("pyjapc")


##########################################


def parse_timestamp(thistime):
    accepted_time_input_format = [
        "%Y-%m-%d %H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d_%H:%M:%S.%f",
        "%Y-%m-%d_%H:%M:%S",
        "%Y/%m/%d %H:%M:%S.%f",
    ]
    for fmat in accepted_time_input_format:
        try:
            dtobject = dt.datetime.strptime(thistime, fmat)
            return dtobject
        except ValueError:
            pass
    timefmatstring = ""
    for fmat in accepted_time_input_format:
        timefmatstring = timefmatstring + '"' + fmat + '" ,   '
    sys.tracebacklimit = 0
    raise ValueError(
        "No appropriate input format found for start time of scan (-s).\n "
        "---> Accepted input formats are:   " + timefmatstring
    )


##########################################


# function to help write output from datetime objects in standard format throughout code
def convert_to_data_output_format(dtobject):
    output_timestamp = dtobject.strftime(formats.TIME)
    return output_timestamp


##########################################


if __name__ == '__main__':
    # Create a PyJapc instance with selector SCT.USER.ALL
    # INCA is automatically configured based on the timing domain you specify here

    CycleName = "LHC.USER.ALL"
    INCAacc = "LHC"
    noSetFlag = True

    japc = pyjapc.PyJapc(selector=CycleName, incaAcceleratorName=INCAacc, noSet=noSetFlag)
    japc.rbacLogin()
    acquesitions_per_file = 100
    j = 0
    t = 0
    while True:
        time.sleep(0.7)
        print(j, t)
        B1_image = japc.getParam("LHC.BSRTS.5R4.B1/Image")
        B2_image = japc.getParam("LHC.BSRTS.5L4.B2/Image")
        if t == 0:
            allB1data = []
            allB2data = []
            B1_IMGtime = B1_image["acqTime"]
            B2_IMGtime = B2_image["acqTime"]
            B1_IMGtime_dt = parse_timestamp(B1_IMGtime)
            B2_IMGtime_dt = parse_timestamp(B2_IMGtime)
            B1_IMGtime_st = convert_to_data_output_format(B1_IMGtime_dt)
            B2_IMGtime_st = convert_to_data_output_format(B2_IMGtime_dt)

        allB1data.append(B1_image)
        allB2data.append(B2_image)
        t += 1
        if t == acquesitions_per_file:
            j += 1
            f1name = "data_BSRT_B1_" + B1_IMGtime_st + ".dat"
            f2name = "data_BSRT_B2_" + B2_IMGtime_st + ".dat"
            f1 = open(f1name, "wb")
            f2 = open(f2name, "wb")
            pickle.dump(allB1data, f1)
            pickle.dump(allB2data, f2)
            f1.close()
            f2.close()
            os.system("gzip " + f1name)
            os.system("gzip " + f2name)
            t = 0

    # Close the RBAC session
    japc.rbacLogout()  # this code in not reachable! Encapsulate in a try..except maybe?



"""
PyLSA
-----

This module provides useful functions to conveniently wrap the functionality of ``pjlsa``.
"""
import jpype
import logging
import re
import tfs
from jpype import java, JException
from omc3.utils.mock import cern_network_import
from omc3.utils.time_tools import AccDatetime
from typing import Callable, Union, Dict, Tuple, List

LOG = logging.getLogger(__name__)
pytimber = cern_network_import("pytimber")
pjlsa = cern_network_import("pjlsa")

try:
    pjLSAClient = pjlsa.LSAClient
except ImportError:
    pjLSAClient = object

RELEVANT_BP_CONTEXTS = ("OPERATIONAL", "MD")
RELEVANT_BP_CATEGORIES = ("DISCRETE",)

HEAD_KNOB = "Knob"
HEAD_OPTICS = "Optics"
HEAD_INFO = "Info"
COL_NAME = "NAME"
COL_CIRCUIT = "CIRCUIT"
PREF_DELTA = "DELTA_"

MAX_RETRIES = 10


class LSAClient(pjLSAClient):
    """Extension of the LSAClient."""

    def __getattr__(self, item):
        """ Overwrite __getattr__ to raise the proper import errors at the proper time."""
        try:
            super().__getattr__(item)
        except AttributeError as e:
            # Hint: we might also end up here, if a function is called on the
            # client, that does not exist. Check the stacktrace!
            pjlsa.pjLSAClient  # might raise the Mock-Class import error
            raise e  # if that worked, raise the actual attribute error

    def find_knob_names(self, accelerator: str = "lhc", regexp: str = "") -> list:
        """
        Return knobs for accelerator.

        Args:
            accelerator: Accelerator name.
            regexp: Regular Expression to filter knob names.

        Returns:
            Sorted list of knob names.
        """
        req = self._ParametersRequestBuilder()
        req.setAccelerator(self._getAccelerator(accelerator))
        req.setParameterTypeName("KNOB")
        lst = self._parameterService.findParameters(req.build())
        reg = re.compile(regexp, re.IGNORECASE)
        return sorted(filter(reg.search, [pp.getName() for pp in lst]))

    def find_existing_knobs(self, knobs: List[str]) -> List[str]:
        """
        Return only the knobs that exist from the given list.
        This function was created out of the need to filter these first,
        as knobs that exist but do not belong to a beamprocess return noting in
        _getTrimsByBeamprocess, while knobs that do not exist at all crashed pjlsa.
        This filter should probably have been in pjlsa's _buildParameterList.

        Args:
            knobs (list): List of strings of the knobs to check.

        Returns:
            A list of the knob names that actually exist.
        """
        dont_exist = [k for k in knobs if self._getParameter(k) is None]
        if len(dont_exist):
            LOG.warning(f"The following knobs do not exist and will be filtered: {dont_exist}.")
            knobs = [k for k in knobs if k not in dont_exist]
        return knobs

    def find_last_fill(
            self, acc_time: AccDatetime, accelerator: str = "lhc", source: str = "nxcals"
    ) -> Tuple[str, list]:
        """
        Return last fill name and BeamProcesses.

         Args:
            acc_time (AccDatetime): Accelerator datetime object.
            accelerator (str): Name of the accelerator.
            source (str): pytimber source

        Returns:
            tuple: Last fill name (str), Beamprocesses of last fill (list).
         """
        start_time = acc_time.sub(days=1)  # assumes a fill is not longer than a day
        try:
            fills = self.find_beamprocess_history(
                t_start=start_time, t_end=acc_time,
                accelerator=accelerator,
                source=source,
            )
        except TypeError as e:
            raise ValueError(
                f"No beamprocesses found in the day before {acc_time.cern_utc_string()}"
            ) from e
        last_fill = sorted(fills.keys())[-1]
        return last_fill, fills[last_fill]

    def find_beamprocess_history(
            self, t_start: AccDatetime, t_end: AccDatetime, accelerator: str = "lhc", source: str = "nxcals"
    ) -> Dict:
        """
        Finds the BeamProcesses between t_start and t_end and sorts then by fills.
        Adapted from pjlsa's FindBeamProcessHistory but with source pass-through
        and trial-loop.

        Args:
            t_start (AccDatetime): start time
            t_end (AccDatetime): end time
            accelerator (str): Name of the accelerator.
            source (str): pytimber source

        Returns:
            Dictionary of fills (keys) with a list of Timestamps and BeamProcesses.

        """
        cts = self.findUserContextMappingHistory(t_start.timestamp(), t_end.timestamp(), accelerator=accelerator)

        db = pytimber.LoggingDB(source=source)
        fillnts, fillnv = try_to_acquire_data(
            db.get, "HX:FILLN", t_start.timestamp(), t_end.timestamp()
        )["HX:FILLN"]

        if not len(fillnv):
            raise ValueError(f"No beamprocesses for {accelerator} ({source}) found between {t_start} - {t_end}.")

        fills = {}
        for ts, name in zip(cts.timestamp, cts.name):
            idx = fillnts.searchsorted(ts) - 1
            filln = int(fillnv[idx])
            fills.setdefault(filln, []).insert(0, (ts, name))
        return fills

    def get_trim_history(
            self, beamprocess: str, knobs: list, 
            start_time: AccDatetime = None, end_time: AccDatetime = None, 
            accelerator: str = "lhc"           
    ) -> dict:
        """
        Get trim history for knobs between specified times.
        If any of the times are not given, all available data in that time-direction
        is extracted.

        Args:
            beamprocess (str): Name of the beamprocess.
            knobs (list): List of strings of the knobs to check.
            start_time (AccDatetime): Accelerator datetime object.
            end_time (AccDatetime): Accelerator datetime object.
            accelerator (str): Name of the accelerator.

        Returns:
            Dictionary of trims and their data (as TrimTuples, i.e. NamedTuple of lists of time and data).
        """
        if knobs is None or len(knobs) == 0:
            knobs = self.find_knob_names(accelerator)
        else:
            knobs = self.find_existing_knobs(knobs)
            if not len(knobs):
                raise ValueError("None of the given knobs exist!")

        if start_time is not None:
            start_time = start_time.timestamp() 
        
        if end_time is not None:
            end_time = end_time.timestamp() 

        try:
            trims = self.getTrims(parameter=knobs, beamprocess=beamprocess, start=start_time, end=end_time)
        except jpype.java.lang.NullPointerException as e:
            # In the past this happened, when a knob was not defined, but
            # this should have been caught by the filter_existing_knobs above
            raise ValueError(f"Something went wrong when extracting trims for the knobs: {knobs}") from e

        trims_not_found = [k for k in knobs if k not in trims.keys()]
        if len(trims_not_found):
            LOG.warning(f"The following knobs were not found in '{beamprocess}': {trims_not_found}")
        return trims


    def get_beamprocess_info(self, beamprocess: Union[str, object]) -> Dict:
        """
        Get context info of the given beamprocess.

        Args:
            beamprocess (str, object): Name of the beamprocess or Beamprocess object.

        Returns:
            Dictionary with context info.
        """
        if isinstance(beamprocess, str):
            beamprocess = self._contextService.findStandAloneBeamProcess(beamprocess)
        bp_dict = _beamprocess_to_dict(beamprocess)
        LOG.debug(str(bp_dict))
        return bp_dict

    def find_active_beamprocess_at_time(
            self, acc_time: AccDatetime, accelerator: str = "lhc",
            bp_group: str = "POWERCONVERTERS"  # the Beamprocesses relevant for OMC,
    ):
        """
        Find the active beam process at the time given.
        Same as what online model extractor (KnobExtractor) does.

        Args:
            acc_time (AccDatetime): Accelerator datetime object.
            accelerator (str): Name of the accelerator.
            bp_group (str): BeamProcess Group, choices : 'POWERCONVERTERS', 'ADT', 'KICKERS', 'SPOOLS', 'COLLIMATORS'


        Returns:
            'cern.lsa.domain.settings.spi.StandAloneBeamProcessImpl'
        """
        if accelerator != "lhc":
            raise NotImplementedError("Active-Beamprocess retrieval is only implemented for LHC")

        beamprocessmap = self._lhcService.findResidentStandAloneBeamProcessesByTime(
            int(acc_time.timestamp() * 1000)  # java timestamps are in milliseconds
        )
        beamprocess = beamprocessmap.get(bp_group)
        if beamprocess is None:
            raise ValueError(f"No active BeamProcess found for group '{bp_group}' "
                             f"at time {acc_time.utc_string} UTC.")
        LOG.debug(f"Active Beamprocess at time '{acc_time.cern_utc_string()}': {str(beamprocess)}")
        return beamprocess

    def get_knob_circuits(self, knob_name: str, optics: str) -> tfs.TfsDataFrame:
        """
        Get a dataframe of the structure of the knob. Similar to online model extractor
        (KnobExtractor.getKnobHiercarchy)
        
        Args:
            knob_name: name of the knob.
            optics: name of the optics.

        Returns:
            A `TfsDataFrame` of the knob circuits.
        """
        LOG.debug(f"Getting knob defintions for '{knob_name}', optics '{optics}'")
        df = tfs.TfsDataFrame()
        df.headers[HEAD_KNOB] = knob_name
        df.headers[HEAD_OPTICS] = optics
        df.headers[HEAD_INFO] = "In MAD-X it should be 'name = name + DELTA * knobValue'"
        knob = self._knobService.findKnob(knob_name)
        if knob is None:
            raise IOError(f"Knob '{knob_name}' does not exist")
        try:
            knob_settings = knob.getKnobFactors().getFactorsForOptic(optics)
        except jpype.java.lang.IllegalArgumentException:
            raise IOError(f"Knob '{knob_name}' not available for optics '{optics}'")

        for knob_factor in knob_settings:
            factor = knob_factor.getFactor()
            circuit = knob_factor.getComponentName()
            param = self._parameterService.findParameterByName(circuit)
            type_ = param.getParameterType().getName()
            madx_name = self.get_madx_name_from_circuit(circuit)
            if madx_name is None:
                LOG.error(
                    f"Circuit '{circuit}' could not be resolved to a MADX name in LSA! "
                    "It will not be found in knob-definition!"
                )
            else:
                LOG.debug(f"  Found component '{circuit}': {madx_name}, {factor}")
                df.loc[madx_name, COL_CIRCUIT] = circuit
                df.loc[madx_name, f"{PREF_DELTA}{type_.upper()}"] = factor
        return df.fillna(0)

    def get_madx_name_from_circuit(self, circuit: str):
        """Returns the ``MAD-X`` Strength Name (Circuit/Knob) from the given circuit name."""
        logical_name = circuit.split("/")[0]
        slist = jpype.java.util.Collections.singletonList(
            logical_name
        )  # python lists did not work (jdilly)
        madx_name = self._deviceService.findMadStrengthNamesByLogicalNames(slist)  # returns a map
        return madx_name[logical_name]


# Single Instance LSAClient ####################################################


class LSAMeta(type):
    """Metaclass for single instance LSAClient."""

    _client = None

    def __getattr__(cls, attr):
        if cls._client is None:
            LOG.debug("Creating LSA Client (only once).")
            cls._client = LSAClient()

        client_attr = cls._client.__getattribute__(attr)
        if callable(client_attr):

            def hooked(*args, **kwargs):
                result = client_attr(*args, **kwargs)
                result_is_self = False
                try:
                    if result == cls._client:
                        # prevent client from becoming unwrapped
                        return cls
                except (ValueError, SystemError):
                    # happens with e.g. numpy arrays (ValueError)
                    # or JavaObjects (SytemError) as return values
                    pass
                return result

            return hooked
        else:
            return client_attr


class LSA(metaclass=LSAMeta):
    """Import this class to use LSA like the client without the need to instantiate it.
    Disadvantage: It will always use the default Server.
    """
    pass


# Helper Functions #############################################################


def _beamprocess_to_dict(bp):
    """Converts some fields of the beamprocess (java) to a dictionary."""
    bp_dict = {'Name': bp.toString(), "Object": bp}
    bp_dict.update({getter[3:]: str(bp.__getattribute__(getter)())  # __getattr__ does not exist
                    for getter in dir(bp)
                    if getter.startswith('get') and "Attribute" not in getter})
    return bp_dict


def try_to_acquire_data(function: Callable, *args, **kwargs):
    """Tries to get data from function multiple times.
    TODO: Move to omc3 as is also used there in BBQ extraction.

     Args:
         function (Callable): function to be called, e.g. db.get
         args, kwargs: arguments passed to ``function``

    Returns:
        Return arguments of ``function``

     """
    retries = MAX_RETRIES
    for tries in range(retries + 1):
        try:
            return function(*args, **kwargs)
        except java.lang.IllegalStateException as e:
            raise IOError("Could not acquire data, user probably has no access to NXCALS") from e
        except JException as e:  # Might be a case for retries
            if "RetryableException" in str(e) and (tries + 1) < retries:
                LOG.warning(f"Could not acquire data! Trial no {tries + 1} / {retries}")
                continue  # will go to the next iteratoin of the loop, so retry
            raise IOError("Could not acquire data!") from e



"""
Constants: Forced DA Analysis
-----------------------------

Specific constants relating to the forced DA analysis to be used in ``PyLHC``, to help with
consistency.
"""
from pylhc.constants.general import PLANE_TO_HV, TFS_SUFFIX

RESULTS_DIR = "forced_da_analysis"

ROLLING_AVERAGE_WINDOW = 100
OUTLIER_LIMIT = 0.5 * 1e-6
TIME_AROUND_KICKS_MIN = 10
TIME_BEFORE_KICK_S = [30, 5]
TIME_AFTER_KICK_S = [5, 30]
YPAD = 0.05  # additional padding of the y axis for DA plots

INITIAL_DA_FIT = 12  # initial DA for fitting in values of nominal emittance
MAX_CURVEFIT_FEV = 10000  # Max number of curve_fit iterations


BWS_DIRECTIONS = ("IN", "OUT")


# Kick File Definitions --------------------------------------------------------

KICKFILE = "kick"


def outfile_kick(plane) -> str:
    return f"{KICKFILE}_fda_{plane.lower()}{TFS_SUFFIX}"


def outfile_emittance(plane) -> str:
    return f"emittance_{plane.lower()}{TFS_SUFFIX}"


def outfile_emittance_bws(plane) -> str:
    return f"emittance_bws_{plane.lower()}{TFS_SUFFIX}"


OUTFILE_INTENSITY = f"intensity{TFS_SUFFIX}"

# Plotting ---------------------------------------------------------------------

PLOT_FILETYPES = (".pdf", ".png")


def outfile_plot(ptype, plane, ftype) -> str:
    return f"{ptype}_{plane.lower()}{ftype}"


# Timber Keys ------------------------------------------------------------------

INTENSITY_KEY = "LHC.BCTFR.A6R4.B{beam:d}:BEAM_INTENSITY"

BSRT_EMITTANCE_SIGMA_FIT_KEY = "LHC.BSRT.5{side:s}4.B{beam:d}:FIT_SIGMA_{plane:s}"
BSRT_EMITTANCE_AVERAGE_KEY = "LHC.BSRT.5{side:s}4.B{beam:d}:AVERAGE_EMITTANCE_{plane:s}"
BSRT_EMITTANCE_TO_METER = 1e-6  # Emittance is normalized and in um

BWS_EMITTANCE_KEY = "LHC.BWS.5{side:s}4.B{beam:d}{plane:s}.APP.{direction:s}:EMITTANCE_NORM"
BWS_EMITTANCE_TO_METER = 1e-6  # Emittance is normalized and in um


LR_MAP = {1: "R", 2: "L"}


def bsrt_emittance_key(beam, plane, type_):
    key = {"fit_sigma": BSRT_EMITTANCE_SIGMA_FIT_KEY, "average": BSRT_EMITTANCE_AVERAGE_KEY}[type_]
    return key.format(side=LR_MAP[beam], beam=beam, plane=PLANE_TO_HV[plane])


def bws_emittance_key(beam, plane, direction) -> str:
    return BWS_EMITTANCE_KEY.format(
        side=LR_MAP[beam], beam=beam, plane=PLANE_TO_HV[plane], direction=direction
    )


# Headers ----------------------------------------------------------------------
HEADER_TIME_BEFORE = "Timespan_before_kick[s]"
HEADER_TIME_AFTER = "Timespan_after_kick[s]"
HEADER_BSRT_ROLLING_WINDOW = "Emittance_rolling_window_length"
HEADER_BSRT_OUTLIER_LIMIT = "Emittance_outlier_limit"
HEADER_DA = "Forced_DA_J_Fit_{plane:}[{unit:}]"
HEADER_DA_ERROR = "Forced_DA_J_Fit_Error_{plane:}[{unit:}]"
HEADER_ENERGY = "Beam_Energy[GeV]"
HEADER_NOMINAL_EMITTANCE = "Nominal_Emittance_{plane:}[m]"


def header_da(plane, unit="m") -> str:
    return HEADER_DA.format(plane=plane.upper(), unit=unit)


def header_da_error(plane, unit="m") -> str:
    return HEADER_DA_ERROR.format(plane=plane.upper(), unit=unit)


def header_nominal_emittance(plane) -> str:
    return HEADER_NOMINAL_EMITTANCE.format(plane=plane.upper())


def header_norm_nominal_emittance(plane) -> str:
    return f"Normalized_{HEADER_NOMINAL_EMITTANCE.format(plane=plane.upper())}"


# Columns ----------------------------------------------------------------------

INTENSITY = "INTENSITY"
INTENSITY_BEFORE = "I_BEFORE"
INTENSITY_AFTER = "I_AFTER"
INTENSITY_LOSSES = "I_LOSSES"

# Column Modifiers ---
CLEAN = "CLEAN"
MEAN = "MEAN"
ERR = "ERR"
RELATIVE = "REL"
SIGMA = "SIGMA"


def err_col(column) -> str:
    return f"{ERR}{column}"


def mean_col(column) -> str:
    return f"{MEAN}{column}"


def rel_col(column) -> str:
    return f"{column}{RELATIVE}"


def sigma_col(column) -> str:
    return f"{SIGMA}{column}"


# Planed columns ---
EMITTANCE = "EMITTANCE"
NORM_EMITTANCE = "NORMEMITTANCE"


def column_action(plane) -> str:
    return f"2J{plane.upper()}RES"


def column_emittance(plane) -> str:
    return f"{EMITTANCE}{plane.upper()}"


def column_norm_emittance(plane) -> str:
    return f"{NORM_EMITTANCE}{plane.upper()}"


def column_bws_norm_emittance(plane, direction) -> str:
    return f"{column_norm_emittance(plane)}_{direction}"



"""
Print Machine Settings Overview
-------------------------------

Prints an overview over the machine settings at a provided given time, or the current settings if
no time is given.
If an output path is given, all info will be written into tfs files,
otherwise a summary is logged into console.

Knob values can be extracted and the knob definition gathered. 
For brevity reasons, this data is not logged into the summary in the console.
If a start time is given, the trim history for the given knobs can be written out as well. 
This data is also not logged.

Can be run from command line, parameters as given in :meth:`pylhc.machine_settings_info.get_info`.
All gathered data is returned, if this function is called from python.

.. code-block:: none

    usage: machine_settings_info.py [-h] [--time TIME] [--start_time START_TIME]
                                    [--knobs KNOBS [KNOBS ...]] [--accel ACCEL]
                                    [--output_dir OUTPUT_DIR] [--knob_definitions]
                                    [--source SOURCE] [--log]

    optional arguments:
    -h, --help            show this help message and exit
    --time TIME           UTC Time as 'Y-m-d H:M:S.f' format.
                          Acts as point in time or end time
                         (if ``start_time`` is given).
    --start_time START_TIME
                            UTC Time as 'Y-m-d H:M:S.f' format.
                            Defines the beginning of the time-range.
    --knobs KNOBS [KNOBS ...]
                            List of knobnames.
    --accel ACCEL         Accelerator name.
    --output_dir OUTPUT_DIR
                            Output directory.
    --knob_definitions    Set to extract knob definitions.
    --source SOURCE       Source to extract data from.
    --log                 Write summary into log
                          (automatically done if no output path is given).


:author: jdilly
"""
from collections import OrderedDict, namedtuple

import tfs
from generic_parser import DotDict, EntryPointParameters, entrypoint
from generic_parser.entry_datatypes import get_instance_faker_meta
from omc3.utils import logging_tools
from omc3.utils.iotools import PathOrStr
from omc3.utils.time_tools import AccDatetime, AcceleratorDatetime
from pathlib import Path
from typing import Tuple, Iterable, Dict, Union

from pylhc.constants import machine_settings_info as const
from pylhc.data_extract.lsa import COL_NAME as LSA_COLUMN_NAME, LSA

LOG = logging_tools.get_logger(__name__)


class AccDatetimeOrStr(metaclass=get_instance_faker_meta(AccDatetime, str)):
    """A class that accepts AccDateTime and strings."""
    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("\'\"")  # behavior like dict-parser, IMPORTANT FOR EVERY STRING-FAKER
        return value


# Main #########################################################################


def _get_params() -> dict:
    """Parse Commandline Arguments and return them as options."""
    return EntryPointParameters(
        time=dict(
            default=None,
            type=AccDatetimeOrStr,
            help=("UTC Time as 'Y-m-d H:M:S.f' format or AccDatetime object."
                  " Acts as point in time or end time (if ``start_time`` is given).")
            ),
        start_time=dict(
            default=None,
            type=AccDatetimeOrStr,
            help=("UTC Time as 'Y-m-d H:M:S.f' format or AccDatetime object."
                  " Defines the beginning of the time-range.")
             ),
        knobs=dict(
            default=None,
            nargs="+",
            type=str,
            help="List of knobnames."),
        accel=dict(
            default='lhc',
            type=str,
            help="Accelerator name."),
        output_dir=dict(
            default=None,
            type=PathOrStr,
            help="Output directory."),
        knob_definitions=dict(
            action="store_true",
            help="Set to extract knob definitions."),
        source=dict(
            type=str,
            default="nxcals",
            help="Source to extract data from."),
        log=dict(
            action="store_true",
            help="Write summary into log (automatically done if no output path is given)."),
    )


@entrypoint(_get_params(), strict=True)
def get_info(opt) -> Dict[str, object]:
    """
    Get info about **Beamprocess**, **Optics** and **Knobs** at given time.

    Keyword Args:

       *--Optional--*

        - **accel** *(str)*:

            Accelerator name.

            default: ``lhc``


        - **knob_definitions**:

            Set to extract knob definitions.

            action: ``store_true``


        - **knobs** *(str)*:

            List of knobnames.

            default: ``None``


        - **log**:

            Write summary into log (automatically done if no output path is
            given).

            action: ``store_true``


        - **output_dir** *(PathOrStr)*:

            Output directory.

            default: ``None``


        - **source** *(str)*:

            Source to extract data from.

            default: ``nxcals``


        - **start_time** *(AccDatetime, str)*:

            UTC Time as 'Y-m-d H:M:S.f' format or AccDatetime object.
            Defines the beginning of the time-range.

            default: ``None``


        - **time** *(AccDatetime, str)*:

            UTC Time as 'Y-m-d H:M:S.f' format or AccDatetime object.
            Acts as point in time or end time (if ``start_time`` is given).

            default: ``None``

    Returns:
        dict: Dictionary containing the given ``time`` and ``start_time``,
        the extracted ``beamprocess``-info and ``optics``-info, the
        ``trim_histories`` and current (i.e. at given ``time``) ``trims``
        and the ``knob_definitions``, if extracted.

    """
    if opt.output_dir is None:
        opt.log = True

    acc_time, acc_start_time = _get_times(opt.time, opt.start_time, opt.accel)
    beamprocess_info = _get_beamprocess(acc_time, opt.accel, opt.source)

    optics_info, knob_definitions, trim_histories, trims = None, None, None, None
    try:
        optics_info = _get_optics(acc_time, beamprocess_info.Name, beamprocess_info.StartTime)
    except ValueError as e:
        LOG.error(str(e))
    else:
        trim_histories = LSA.get_trim_history(beamprocess_info.Object, opt.knobs, start_time=acc_start_time, end_time=acc_time, accelerator=opt.accel)
        trims = _get_last_trim(trim_histories)
        if opt.knob_definitions:
            knob_definitions = _get_knob_definitions(opt.knobs, optics_info.Name)

    if opt.log:
        log_summary(acc_time, beamprocess_info, optics_info, trims)

    if opt.output_dir is not None:
        out_path = Path(opt.output_dir)
        out_path.mkdir(parents=True, exist_ok=True)
        write_summary(out_path, opt.accel, acc_time, beamprocess_info, optics_info, trims)
        
        if trim_histories and acc_start_time:
            write_trim_histories(out_path, trim_histories, opt.accel, acc_time, acc_start_time, beamprocess_info, optics_info)

        if knob_definitions:
            write_knob_defitions(out_path, knob_definitions)

    return {
        "time": acc_time,
        "start_time": acc_start_time,
        "beamprocess": beamprocess_info,
        "optics": optics_info,
        "trim_histories": trim_histories,
        "trims": trims,
        "knob_definitions": knob_definitions
    }


# Output #######################################################################


def log_summary(acc_time: AccDatetime, bp_info: DotDict,
                optics_info: DotDict = None, trims: Dict[str, float] = None):
    """Log the summary.

    Args:
        acc_time (AccDatetime): User given Time
        bp_info (DotDict): BeamProcess Info Dictionary
        optics_info (DotDict): Optics Info Dictionary
        trims (dict): Trims key-value dictionary
    """
    summary = (
        "\n----------- Summary ---------------------\n"
        f"Given Time:   {acc_time.utc_string[:-7]} UTC\n"
        f"Fill:         {bp_info.Fill:d}\n"
        f"Beamprocess:  {bp_info.Name}\n"
        f"  Start:      {bp_info.StartTime.utc_string[:-7]} UTC\n"
        f"  Context:    {bp_info.ContextCategory}\n"
        f"  Descr.:     {bp_info.Description}\n"
    )
    if optics_info is not None:
        summary += (
            f"Optics:       {optics_info.Name}\n"
            f"  Start:      {optics_info.StartTime.utc_string[:-7]} UTC\n"
        )

    if trims is not None:
        summary += "----------- Trims -----------------------\n"
        for trim, value in trims.items():
            summary += f"{trim:30s}: {value:g}\n"

    summary += "-----------------------------------------\n\n"
    LOG.info(summary)


def write_summary(
    output_path: Path, accel: str, acc_time: AccDatetime, bp_info: DotDict,
    optics_info: DotDict = None, trims: Dict[str, float] = None
):
    """Write summary into a ``tfs`` file.

    Args:
        output_path (Path): Folder to write output file into
        accel (str): Name of the accelerator
        acc_time (AccDatetime): User given Time
        bp_info (DotDict): BeamProcess Info Dictionary
        optics_info (DotDict): Optics Info Dictionary
        trims (dict): Trims key-value dictionary
    """
    if trims is not None:
        trims = trims.items()

    info_tfs = tfs.TfsDataFrame(trims, columns=[const.column_knob, const.column_value])
    info_tfs.headers = OrderedDict([
        ("Hint:",                           "All times given in UTC."),
        (const.head_accel,                  accel),
        (const.head_time,                   acc_time.cern_utc_string()),
        (const.head_beamprocess,            bp_info.Name),
        (const.head_fill,                   bp_info.Fill),
        (const.head_beamprocess_start,      bp_info.StartTime.cern_utc_string()),
        (const.head_context_category,       bp_info.ContextCategory),
        (const.head_beamprcess_description, bp_info.Description),
        ])
    if optics_info is not None:
        info_tfs.headers.update(OrderedDict([
            (const.head_optics, optics_info.Name),
            (const.head_optics_start, optics_info.StartTime.cern_utc_string()),
        ]))
    tfs.write(output_path / const.info_name, info_tfs)


def write_knob_defitions(output_path: Path, definitions: dict):
    """Write Knob definitions into a **tfs** file."""
    for knob, definition in definitions.items():
        path = output_path / f"{knob.replace('/', '_')}{const.knobdef_suffix}"
        tfs.write(path, definition, save_index=LSA_COLUMN_NAME)


def write_trim_histories(
    output_path: Path, trim_hisotries: Dict[str, namedtuple], accel: str,
    acc_time: AccDatetime = None, acc_start_time: AccDatetime = None, 
    bp_info: DotDict = None, optics_info: DotDict = None
):
    """ Write the trim histories into tfs files.
    There are two time columns, one with timestamps as they are usually easier to handle
    and one with the UTC-string, as they are more human readable.

    Args:
        output_path (Path): Folder to write output file into
        trim_hisotries (dict): trims histories as extracted via LSA.get_trim_history()
        accel (str): Name of the accelerator
        acc_time (AccDatetime): User given (End)Time
        acc_start_time (AccDatetime): User given Start Time
        bp_info (DotDict): BeamProcess Info Dictionary
        optics_info (DotDict): Optics Info Dictionary
    """
    AccDT = AcceleratorDatetime[accel]

    # Create headers with basic info ---
    headers = OrderedDict([("Hint:", "All times are given in UTC."), 
                            (const.head_accel, accel)
    ])
    
    if acc_start_time:
        headers.update({const.head_start_time: acc_start_time.cern_utc_string()})

    if acc_time:
        headers.update({const.head_end_time: acc_time.cern_utc_string()})
    
    if bp_info:
        headers.update({
            const.head_beamprocess: bp_info.Name, 
            const.head_fill: bp_info.Fill,
        })

    if optics_info:
        headers.update({const.head_optics: optics_info.Name})

    # Write trim history per knob ----
    for knob, trim_history in trim_hisotries.items():
        trims_tfs = tfs.TfsDataFrame(headers=headers, columns=[const.column_time, const.column_timestamp, const.column_value])
        for timestamp, value in zip(trim_history.time, trim_history.data):
            time = AccDT.from_timestamp(timestamp).cern_utc_string()
            try:
                len(value)
            except TypeError:
                # single value (as it should be)
                trims_tfs.loc[len(trims_tfs), :] = (time, timestamp, value)
            else:
                # multiple values (probably weird)
                LOG.debug("Multiple values in trim for {knob} at {time}.")
                for item in value:
                    trims_tfs.loc[len(trims_tfs), :] = (time, timestamp, item)

        path = output_path / f"{knob.replace('/', '_')}{const.trimhistory_suffix}"
        tfs.write(path, trims_tfs)


# Beamprocess ##################################################################


def _get_beamprocess(acc_time: AccDatetime, accel: str, source: str) -> DotDict:
    """Get the info about the active beamprocess at ``acc_time``."""
    fill_no, fill_bps = LSA.find_last_fill(acc_time, accel, source)
    beamprocess = LSA.find_active_beamprocess_at_time(acc_time)
    try:
        start_time = _get_beamprocess_start(fill_bps, acc_time, str(beamprocess))
    except ValueError as e:
        raise ValueError(f"In fill {fill_no} the {str(e)}") from e
    bp_info = LSA.get_beamprocess_info(beamprocess)
    bp_info.update({"Fill": fill_no, "StartTime": start_time})
    return DotDict(bp_info)


def _get_beamprocess_start(beamprocesses: Iterable[Tuple[float, str]], acc_time: AccDatetime, bp_name: str) -> AccDatetime:
    """
    Get the last beamprocess in the list of beamprocesses before dt_utc.
    Returns the start time of the beam-process in utc.
    """
    LOG.debug(f"Looking for beamprocess '{bp_name}' in fill before '{acc_time.cern_utc_string()}'")
    ts = acc_time.timestamp()
    for time, name in sorted(beamprocesses, key=lambda x: x[0], reverse=True):
        if time <= ts and name == bp_name:
            return acc_time.__class__.from_timestamp(time)
    raise ValueError(
        f"Beamprocess '{bp_name}' was not found."
    )


# Optics #######################################################################


def _get_optics(acc_time: AccDatetime, beamprocess: str, bp_start: AccDatetime) -> DotDict:
    """Get the info about the active optics at ``acc_time``."""
    optics_table = LSA.getOpticTable(beamprocess)
    optics, start_time = _get_last_optics(optics_table, beamprocess, bp_start, acc_time)
    return DotDict({"Name": optics, "StartTime": start_time})


def _get_last_optics(
    optics_table, bp: str, bp_start: AccDatetime, acc_time: AccDatetime
) -> (str, AccDatetime):
    """Get the name of the optics at the right time for current beam process."""
    ts = acc_time.timestamp() - bp_start.timestamp()
    item = None
    for item in reversed(optics_table):
        if item.time <= ts:
            break
    if item is None:
        raise ValueError(f"No optics found for beamprocess {bp}")
    return item.name, acc_time.__class__.from_timestamp(item.time + bp_start.timestamp())


# Knobs ########################################################################


def _get_knob_definitions(knobs: list, optics: str):
    """Get knob definitions."""
    defs = {}
    for knob in knobs:
        try:
            defs[knob] = LSA.get_knob_circuits(knob, optics)
        except IOError as e:
            LOG.warning(e.args[0])
    return defs


def _get_last_trim(trims: dict) -> dict:
    """Returns the last trim in the trim history.

    Args:
        trims (dict): trim history as extracted via LSA.get_trim_history()

    Returns:
        Dictionary of knob names and their values.
    """
    trim_dict = {trim: trims[trim].data[-1] for trim in trims.keys()}  # return last set value
    for trim, value in trim_dict.items():
        try:
            trim_dict[trim] = value.flatten()[-1]  # the very last entry ...
        except AttributeError:
            continue  # single value, as expected
        else:
            LOG.warning(f"Trim {trim} hat multiple data entries {value}, taking only the last one.")
    return trim_dict


# Other ########################################################################


def _get_times(time: Union[str, AccDatetime], start_time: Union[str, AccDatetime], accel: str):
    """ Returns acc_time and acc_start_time parameters depending on the user input. """
    acc_dt = AcceleratorDatetime[accel]
    try:
        time = acc_dt.now() if time is None else acc_dt.from_utc_string(time)
    except TypeError:
        pass  # is already AccDatetime object

    try:
        start_time = None if start_time is None else acc_dt.from_utc_string(start_time)
    except TypeError:
        pass  # is already AccDatetime object

    return time, start_time


# Script Mode ##################################################################


if __name__ == "__main__":
    get_info()



"""
Constants: Machine Settings Info
---------------------------------

Specific constants relating to the retrieval of machine settings information to be used in
``PyLHC``, to help with consistency.
"""
from pylhc.constants.general import TFS_SUFFIX

# TFS-File Conventions #########################################################
# Filename
info_name = f"machine_settings{TFS_SUFFIX}"
knobdef_suffix = f"_definition{TFS_SUFFIX}"
trimhistory_suffix = f"_trims{TFS_SUFFIX}"

# Columns
column_knob = "KNOB"
column_time = "TIME"
column_timestamp = "TIMESTAMP"
column_value = "VALUE"


# Headers
head_accel = "Accelerator"
head_time = "Time"
head_start_time = "StartTime"
head_end_time = "EndTime"
head_beamprocess = "Beamprocess"
head_fill = "Fill"
head_beamprocess_start = "BeamprocessStart"
head_context_category = "ContextCategory"
head_beamprcess_description = "Description"
head_optics = "Optics"
head_optics_start = "OpticsStart"



"""
Constants: BPM Calibration
--------------------------

Specific constants related to the BPM calibration script in ``PyLHC``.
"""

# IPS to be used to compute the calibraton factors
IPS = [1, 4, 5]

# Constants for TFS files
LABELS = ['S',
          'CALIBRATION',
          'ERROR_CALIBRATION',
          'CALIBRATION_FIT',
          'ERROR_CALIBRATION_FIT']
TFS_INDEX = 'NAME'
D = 'D'
ND = 'ND'

# Estimation for the curve fit
BETA_STAR_ESTIMATION = 200

# Methods to be used to compulte the calibration factors
METHODS = ('beta', 'dispersion')

# File name prefix for calibration output
# end result example: {'beta': 'calibration_beta_.tfs', 'dispersion' ... }
CALIBRATION_NAME = {m: f'calibration_{m}_' for m in METHODS}

# Define BPMs to be used for a combination of IP and Beam
BPMS = {1: {1: ['BPMR.5L1.B1',
                'BPMYA.4L1.B1',
                'BPMWB.4L1.B1',
                'BPMSY.4L1.B1',
                'BPMS.2L1.B1',
                'BPMSW.1L1.B1',
                'BPMSW.1R1.B1',
                'BPMS.2R1.B1',
                'BPMSY.4R1.B1',
                'BPMWB.4R1.B1',
                'BPMYA.4R1.B1'],
            2: ['BPM.5L1.B2', 
                'BPMYA.4L1.B2', 
                'BPMWB.4L1.B2', 
                'BPMSY.4L1.B2', 
                'BPMS.2L1.B2', 
                'BPMSW.1L1.B2', 
                'BPMSW.1R1.B2', 
                'BPMS.2R1.B2', 
                'BPMSY.4R1.B2', 
                'BPMWB.4R1.B2', 
                'BPMYA.4R1.B2']
            },
        4: {1: [
                'BPMYA.5L4.B1',
                'BPMWI.A5L4.B1',
                'BPMWA.B5L4.B1',
                'BPMWA.A5L4.B1',
                'BPMWA.A5R4.B1',
                'BPMWA.B5R4.B1',
                'BPMYB.5R4.B1',
                'BPMYA.6R4.B1',
                ],
            2: [ 
                'BPMYB.5L4.B2',
                'BPMWA.B5L4.B2',
                'BPMWA.A5L4.B2',
                'BPMWA.A5R4.B2',
                'BPMWA.B5R4.B2',
                'BPMWI.A5R4.B2',
                'BPMYA.5R4.B2',
                'BPMYB.6R4.B2'
                ]
             },
        5: {1: ['BPMYA.4L5.B1', 
                'BPMWB.4L5.B1', 
                'BPMSY.4L5.B1', 
                'BPMS.2L5.B1', 
                'BPMSW.1L5.B1', 
                'BPMSW.1R5.B1', 
                'BPMS.2R5.B1', 
                'BPMSY.4R5.B1',
                'BPMWB.4R5.B1', 
                'BPMYA.4R5.B1', 
                'BPM.5R5.B1'],
            2: ['BPMYA.4L5.B2', 
                'BPMWB.4L5.B2', 
                'BPMSY.4L5.B2', 
                'BPMS.2L5.B2', 
                'BPMSW.1L5.B2', 
                'BPMSW.1R5.B2', 
                'BPMS.2R5.B2', 
                'BPMSY.4R5.B2', 
                'BPMWB.4R5.B2', 
                'BPMYA.4R5.B2', 
                'BPMR.5R5.B2']
            }
        }

# For the dispersion method, only a subject of the BPMs is used
# Same as BPM: IP and then beam
D_BPMS = {1: {1: ['BPMSY.4L1.B1',
                  'BPMS.2L1.B1',
                  'BPMSW.1L1.B1',
                  'BPMSW.1R1.B1',
                  'BPMS.2R1.B1',
                  'BPMSY.4R1.B1'],
              2: ['BPMSY.4L1.B2', 
                  'BPMS.2L1.B2', 
                  'BPMSW.1L1.B2', 
                  'BPMSW.1R1.B2', 
                  'BPMS.2R1.B2', 
                  'BPMSY.4R1.B2']
              },
          4: {1: [],
              2: []
              },
          5: {1: ['BPMSY.4L5.B1', 
                  'BPMS.2L5.B1', 
                  'BPMSW.1L5.B1', 
                  'BPMSW.1R5.B1', 
                  'BPMS.2R5.B1', 
                  'BPMSY.4R5.B1',
                  ],
              2: ['BPMSY.4L5.B2', 
                  'BPMS.2L5.B2', 
                  'BPMSW.1L5.B2', 
                  'BPMSW.1R5.B2', 
                  'BPMS.2R5.B2', 
                  'BPMSY.4R5.B2']
              }
          }



"""
Beta
-----

The functions in this script compute the calibration factors for the LHC BPMs
using the beta method. The `get_calibration_factors_from_beta` is intended
to be used with the script `bpm_calibration.py`.

"""
from pathlib import Path
from typing import Dict, List, Sequence, Tuple

import numpy as np
import pandas as pd
import tfs
from omc3.optics_measurements.constants import AMP_BETA_NAME, BETA, BETA_NAME, ERR, EXT, S
from omc3.utils import logging_tools
from scipy.optimize import curve_fit

from pylhc.constants.calibration import BETA_STAR_ESTIMATION, BPMS, IPS, LABELS, TFS_INDEX
from pylhc.constants.general import PLANES

LOG = logging_tools.get_logger(__name__)


def _get_beta_fit(
    bpms: Sequence[str],
    beta_phase_tfs: pd.DataFrame,
    plane: str,
) -> pd.DataFrame:
    """
    This function returns a fit of the given beta values along with the
    associated error.

    Args:
        bpms (Sequence[str]): Names of the BPMs to use for the fitting.
        beta_phase_tfs (pd.DataFrame): A ``DataFrame`` with beta from phase values to be fitted at the
            given BPMs.
        plane (str): plane to perform the fit on.

    Returns:
        A ``pandas.DataFrame`` with the resulting values from the fit of the input beta values and the
        associated errors, as columns.
    """

    def beta_function(x, a, b):
        return a + ((x - b) ** 2) / a

    def err_function(x, popt, pcov):
        sa, sb, sab = pcov[0, 0], pcov[1, 1], pcov[0, 1]
        a, b = popt[0], popt[1]

        beta_err = ((a ** 2 - (x - b) ** 2) / a ** 2) ** 2 * sa
        beta_err += 4 * ((x - b) / a) ** 2 * sb
        beta_err -= 4 * (x - b) * (a ** 2 - (x - b) ** 2) / a ** 3 * sab
        return beta_err

    positions = beta_phase_tfs.reindex(bpms)[f"{S}"]
    beta_phase = beta_phase_tfs.reindex(bpms)[f"{BETA}{plane}"]
    beta_phase_err = beta_phase_tfs.reindex(bpms)[f"{ERR}{BETA}{plane}"]

    # Get the rough IP position and beta star for the initial values
    ip_position = (positions[-1] - positions[0]) / 2
    initial_values = (BETA_STAR_ESTIMATION, ip_position)

    # Get the curve fit for the expected parabola
    valid = ~(np.isnan(positions) | np.isnan(beta_phase))

    additional_args = {}
    if sum(beta_phase_err[valid]) != 0:
        additional_args = {"sigma": beta_phase_err[valid]}

    popt, pcov = curve_fit(
        beta_function,
        positions[valid],
        beta_phase[valid],
        p0=initial_values,
        maxfev=1000000,
        absolute_sigma=True,
        **additional_args,
    )

    # Get the error from the covariance matrix
    perr = np.sqrt(np.diag(pcov))

    # Get the fitted beta and add the errors to get min/max values
    beta_fit = beta_function(positions[valid], *popt)
    beta_fit_err = err_function(positions[valid], popt, pcov)

    return pd.DataFrame({f"{BETA}{plane}": beta_fit, f"{ERR}{BETA}{plane}": beta_fit_err})


def _get_factors_from_phase(
    beta_phase_tfs: pd.DataFrame,
    beta_amp_tfs: pd.DataFrame,
    plane: str,
) -> pd.DataFrame:
    """
    This function computes the calibration factors for the beta method with the
    beta from phase values. The associated error is also calculated.
    The equations being the same for the factors from phase and phase fit, this
    function can be used for both.

    Args:
        beta_phase_tfs (pd.DataFrame): A ``DataFrame`` with beta from phase values.
        beta_amp_tfs (pd.DataFrame): A ``DataFrame`` with beta from amplitude values.
        plane (str): Plane of the BPM measurements.

    Returns:
        A ``pandas.DataFrame`` with the computed calibration factors and their associated errors as columns.
    """
    beta_phase = beta_phase_tfs[f"{BETA}{plane}"]
    beta_phase_err = beta_phase_tfs[f"{ERR}{BETA}{plane}"]
    beta_amp = beta_amp_tfs[f"{BETA}{plane}"]
    beta_amp_err = beta_amp_tfs[f"{ERR}{BETA}{plane}"]

    # Compute the calibration factors
    factors = np.sqrt(beta_phase / beta_amp)

    # Now compute the errors
    calibration_error = (beta_phase_err ** 2) / (4 * beta_amp * beta_phase)
    calibration_error += (beta_phase * (beta_amp_err ** 2)) / (4 * (beta_amp ** 3))
    calibration_error = np.sqrt(calibration_error)

    return pd.DataFrame({LABELS[1]: factors, LABELS[2]: calibration_error})


def _get_factors_from_phase_fit(
    beta_phase_tfs: pd.DataFrame,
    beta_amp_tfs: pd.DataFrame,
    ips: Sequence[int],
    plane: str,
) -> Tuple[pd.Series, pd.Series]:
    """
    This function computes the calibration factors for the beta method with the
    beta from phase fit values. The associated error is also calculated.
    This function is used as a wrapper on the `_get_factors_from_fit` function,
    mainly to loop over IPs and filter the BPM list.

    Args:
        beta_phase_tfs (pd.DataFrame): A ``DataFrame`` with beta from phase values.
        beta_amp_tfs (pd.DataFrame): A ``DataFrame`` with beta from amplitude values.
        ips (Sequence[int]): List of IPs to compute the factors for.
        plane (str): Plane of the BPM measurements.

    Returns:
      Tuple[pd.Series, pd.Series]: The first Series are the calibration
      factors, the second one their error.
    """
    # Get the beam concerned by those tfs files
    beam = int(beta_phase_tfs.iloc[0].name[-1])

    calibration_phase_fit, calibration_phase_fit_err = None, None
    for ip in ips:
        LOG.info(f"    Computing the calibration factors from phase fit for IP {ip}")

        # Check for possible missing bpms
        bpms = beta_phase_tfs.reindex(BPMS[ip][beam])
        bpms_amp = beta_amp_tfs.reindex(BPMS[ip][beam])
        for bpm_set in [bpms, bpms_amp]:
            missing = set(bpm_set.loc[bpm_set.isnull().values].index)
            if missing:
                LOG.warning(
                    "    One or several BPMs are missing in the input"
                    " DataFrame, the calibration factors calculation"
                    f" from fit may not be accurate: {missing}"
                )

        # Curve fit the beta from phase values
        beta_phase_fit = _get_beta_fit(BPMS[ip][beam], beta_phase_tfs, plane)

        # Get the factors and put them all together to have all ips in one
        # Series
        c_fit = _get_factors_from_phase(beta_phase_fit, beta_amp_tfs.reindex(BPMS[ip][beam]), plane)
        if calibration_phase_fit is None:
            calibration_phase_fit = c_fit
        else:
            calibration_phase_fit = calibration_phase_fit.append(c_fit)

    # Change the colum names for _fit
    calibration_phase_fit.columns = (LABELS[3], LABELS[4])

    return calibration_phase_fit


def get_calibration_factors_from_beta(ips: Sequence[int], input_path: Path) -> Dict[str, pd.DataFrame]:
    """
    This function is the main function to compute the calibration factors for
    the beta method.
    Given a list of IPs and the path containing the corresponding Tfs files,
    this function returns the calibration factors using both the beta from
    phase and its fitted values.

    Args:
      ips (Sequence[int]): IPs to compute the calibration factors for.
      input_path (Path): Path of the directory containing the beta files.

    Returns:
       Dict[str, pd.DataFrame]: The returned DataFrame object contains the
       calibration factors for each BPM along with their error. Both the beta
       from phase and beta from phase fitted values are used, resulting in 6
       columns:

         - NAME: BPM Name
         - S: Position
         - CALIBRATION: Calibration factors computed from beta from phase
         - ERROR_CALIBRATION: Associated error to the above calibration
           factors
         - CALIBRATION_FIT: Calibration factor computed from fitted beta
           from phase
         - ERROR_CALIBRATION_FIT: Associated error to the above calibration
           factors

    """
    LOG.info("Computing the calibration factors via beta")
    # Loop over each plane and compute the calibration factors
    calibration_factors = dict()
    for plane in PLANES:
        LOG.info(f"  Computing the calibration factors for plane {plane}")

        # Load the tfs files for beta from phase and beta from amp
        beta_phase_tfs = tfs.read(input_path / f"{BETA_NAME}{plane.lower()}{EXT}", index=TFS_INDEX)
        beta_amp_tfs = tfs.read(input_path / f"{AMP_BETA_NAME}{plane.lower()}{EXT}", index=TFS_INDEX)

        # Get the calibration factors from phase
        calibration_phase = _get_factors_from_phase(beta_phase_tfs, beta_amp_tfs, plane)

        # Calibration from phase fit can only be obtained via ballistic optics
        if ips is not None:
            calibration_phase_fit = _get_factors_from_phase_fit(beta_phase_tfs, beta_amp_tfs, ips, plane)
        else:
            calibration_phase_fit = pd.DataFrame(columns=(LABELS[3], LABELS[4]))

        # Assemble the calibration factors in one dataframe
        factors = pd.concat([beta_phase_tfs[S], calibration_phase, calibration_phase_fit], axis=1)
        factors.columns = LABELS
        factors.index.name = TFS_INDEX
        calibration_factors[plane] = factors

    return calibration_factors



"""
Forced DA Analysis
------------------

Top-level script to run the forced DA analysis, following the procedure described in
`CarlierForcedDA2019`_.

Arguments:

*--Required--*

- **beam** *(int)*: Beam to use.

  Flags: **['-b', '--beam']**
  Choices: ``[1, 2]``
- **energy** *(MultiClass)*: Beam energy in GeV.

  Flags: **['-e', '--energy']**
- **kick_directory** *(MultiClass)*: Analysis kick_directory containing kick files.

  Flags: **['-k', '--kickdir']**
- **plane** *(str)*: Plane of the kicks.

  Flags: **['-p', '--plane']**
  Choices: ``['X', 'Y']``

*--Optional--*

- **emittance_outlier_limit** *(float)*: Limit, i.e. cut from mean, on emittance outliers in meter.

  Default: ``5e-07``
- **emittance_tfs** *(MultiClass)*: Dataframe or Path of pre-saved emittance tfs.

- **emittance_type** *(str)*: Which BSRT data to use (from database).

  Choices: ``['fit_sigma', 'average']``
  Default: ``average``
- **emittance_window_length** *(int)*: Length of the moving average window. (# data points)

  Default: ``100``
- **fill** *(int)*: Fill that was used. If not given, check out time_around_kicks.

  Flags: **['-f', '--fill']**
- **fit** *(str)*: Fitting function to use (rearranges parameters to make sense).

  Choices: ``['exponential', 'linear']``
  Default: ``exponential``
- **intensity_tfs** *(MultiClass)*: Dataframe or Path of pre-saved intensity tfs.

- **intensity_time_after_kick** *(int)*: Defines the times after the kicks (in seconds) which is used for intensity averaging to calculate the losses.

  Default: ``[5, 30]``
- **intensity_time_before_kick** *(int)*: Defines the times before the kicks (in seconds) which is used for intensity averaging to calculate the losses.

  Default: ``[30, 5]``
- **normalized_emittance** *(float)*: Assumed NORMALIZED nominal emittance for the machine.

  Default: ``3.7499999999999997e-06``
- **output_directory** *(MultiClass)*: Output kick_directory, if not given subfolder in kick kick_directory

  Flags: **['-o', '--outdir']**
- **pagestore_db** *(MultiClass)*: (Path to-) presaved timber database

- **show**: Show plots.

  Action: ``store_true``
- **show_wirescan_emittance** *(BoolOrPathOrDataFrame)*: Flag if the emittance from wirescan should also be shown, can also be a Dataframe or Path of pre-saved emittance bws tfs.

  Default: ``False``
- **timber_db** *(str)*: Which timber database to use.

  Choices: ``['all', 'mdb', 'ldb', 'nxcals']``
  Default: ``all``
- **time_around_kicks** *(int)*: If no fill is given, this defines the time (in minutes) when data before the first and after the last kick is extracted.

  Default: ``10``
- **plot_styles** *(str)*: Which plotting styles to use,
  either from omc3 styles or default mpl.

  Default: ``['standard']``
- **manual_style** *(DictAsString)*: Additional style rcParameters which update the set of predefined ones.

  Default: ``{}``


:author: jdilly

.. _CarlierForcedDA2019: https://journals.aps.org/prab/pdf/10.1103/PhysRevAccelBeams.22.031002
"""
import os
from collections import defaultdict
from contextlib import suppress
from pathlib import Path
from typing import Tuple

import matplotlib as mpl
import matplotlib.colors as mcolors
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib.transforms as mtrans
import numpy as np
import pandas as pd
import scipy.odr
import scipy.optimize
import tfs
from generic_parser import EntryPointParameters, entrypoint
from generic_parser.entry_datatypes import (
    DictAsString,
    FALSE_ITEMS,
    TRUE_ITEMS,
    get_instance_faker_meta,
    get_multi_class,
)
from generic_parser.tools import DotDict
from omc3.optics_measurements import toolbox
from omc3.plotting.utils import annotations, colors, lines, style
from omc3.tune_analysis.bbq_tools import clean_outliers_moving_average
from omc3.utils import logging_tools
from omc3.utils.iotools import save_config
from omc3.utils.mock import cern_network_import
from omc3.utils.time_tools import CERNDatetime
from pandas import DataFrame, Series
from pandas.plotting import register_matplotlib_converters
from tfs import TfsDataFrame
from tfs.tools import significant_digits

pytimber = cern_network_import('pytimber')
PageStore = cern_network_import('pytimber.pagestore.PageStore')


from pylhc.constants.forced_da_analysis import (
    BSRT_EMITTANCE_TO_METER,
    BWS_DIRECTIONS,
    BWS_EMITTANCE_TO_METER,
    HEADER_BSRT_OUTLIER_LIMIT,
    HEADER_BSRT_ROLLING_WINDOW,
    HEADER_ENERGY,
    HEADER_TIME_AFTER,
    HEADER_TIME_BEFORE,
    INITIAL_DA_FIT,
    INTENSITY,
    INTENSITY_AFTER,
    INTENSITY_BEFORE,
    INTENSITY_KEY,
    INTENSITY_LOSSES,
    KICKFILE,
    MAX_CURVEFIT_FEV,
    OUTFILE_INTENSITY,
    OUTLIER_LIMIT,
    PLOT_FILETYPES,
    RESULTS_DIR,
    ROLLING_AVERAGE_WINDOW,
    TIME_AFTER_KICK_S,
    TIME_AROUND_KICKS_MIN,
    TIME_BEFORE_KICK_S,
    YPAD,
    bsrt_emittance_key,
    bws_emittance_key,
    column_action,
    column_bws_norm_emittance,
    column_emittance,
    column_norm_emittance,
    err_col,
    header_da,
    header_da_error,
    header_nominal_emittance,
    header_norm_nominal_emittance,
    mean_col,
    outfile_emittance,
    outfile_emittance_bws,
    outfile_kick,
    outfile_plot,
    rel_col,
    sigma_col,
)
from pylhc.constants.general import (
    LHC_NOMINAL_EMITTANCE,
    TFS_SUFFIX,
    TIME_COLUMN,
    get_proton_beta,
    get_proton_gamma,
)

LOG = logging_tools.get_logger(__name__)


# Weird Datatypes
class BoolOrPathOrDataFrame(
    metaclass=get_instance_faker_meta(bool, Path, str, tfs.TfsDataFrame, pd.DataFrame, type(None))
):
    """
    A class that behaves like a `boolean` when possible, otherwise like a `Path`, `string` or
    `Dataframe`.
    """

    def __new__(cls, value):
        if isinstance(value, str):
            value = value.strip("'\"")  # behavior like dict-parser

        if value in TRUE_ITEMS:
            return True

        elif value in FALSE_ITEMS:
            return False

        else:
            try:
                return Path(value)
            except TypeError:
                return value


def _get_pathclass(*other_classes):
    class SomethingOrPath(metaclass=get_instance_faker_meta(Path, str, *other_classes, type(None))):
        """A class that behaves like a if possible `Path`, `string` or something else."""

        def __new__(cls, value):
            if isinstance(value, str):
                value = value.strip("'\"")  # Needs to be done for strings in config-files

            try:
                return Path(value)
            except TypeError:
                return value

    return SomethingOrPath


PathOrDataframe = _get_pathclass(tfs.TfsDataFrame, pd.DataFrame)
PathOrPagestore = _get_pathclass(PageStore)
PathOrString = _get_pathclass()


def get_params():
    return EntryPointParameters(
        kick_directory=dict(
            flags=["-k", "--kickdir"],
            required=True,
            type=PathOrString,
            help="Analysis kick_directory containing kick files.",
        ),
        output_directory=dict(
            flags=["-o", "--outdir"],
            type=PathOrString,
            help="Output kick_directory, if not given subfolder in kick kick_directory",
        ),
        energy=dict(
            flags=["-e", "--energy"],
            required=True,
            type=get_multi_class(float, int),
            help="Beam energy in GeV.",
        ),
        fill=dict(
            flags=["-f", "--fill"],
            type=get_multi_class(int, type(None)),
            help="Fill that was used. If not given, check out time_around_kicks.",
        ),
        beam=dict(
            flags=["-b", "--beam"], required=True, choices=[1, 2], type=int, help="Beam to use."
        ),
        plane=dict(
            flags=["-p", "--plane"],
            choices=["X", "Y"],
            required=True,
            type=str,
            help=(
                "Plane of the kicks."
                # " Give 'XY' for using both planes (e.g. diagonal kicks)."  # Future release
            ),
        ),
        time_around_kicks=dict(
            type=int,
            default=TIME_AROUND_KICKS_MIN,
            help=(
                "If no fill is given, this defines the time (in minutes) "
                "when data before the first and after the last kick is extracted."
            ),
        ),
        intensity_time_before_kick=dict(
            type=int,
            nargs=2,
            default=TIME_BEFORE_KICK_S,
            help=(
                "Defines the times before the kicks (in seconds) "
                "which is used for intensity averaging to calculate the losses."
            ),
        ),
        intensity_time_after_kick=dict(
            type=int,
            nargs=2,
            default=TIME_AFTER_KICK_S,
            help=(
                "Defines the times after the kicks (in seconds) "
                "which is used for intensity averaging to calculate the losses."
            ),
        ),
        normalized_emittance=dict(
            type=float,
            default=LHC_NOMINAL_EMITTANCE,
            help="Assumed NORMALIZED nominal emittance for the machine.",
        ),
        emittance_tfs=dict(
            type=PathOrDataframe, help="Dataframe or Path of pre-saved emittance tfs.",
        ),
        intensity_tfs=dict(
            type=PathOrDataframe, help="Dataframe or Path of pre-saved intensity tfs.",
        ),
        show_wirescan_emittance=dict(
            default=False,
            type=BoolOrPathOrDataFrame,
            help=(
                "Flag if the emittance from wirescan should also be shown, "
                "can also be a Dataframe or Path of pre-saved emittance bws tfs."
            ),
        ),
        timber_db=dict(
            type=str,
            default="all",
            choices=["all", "mdb", "ldb", "nxcals"],
            help="Which timber database to use.",
        ),
        pagestore_db=dict(type=PathOrPagestore, help="(Path to-) presaved timber database"),
        fit=dict(
            type=str,
            default="exponential",
            choices=["exponential", "linear"],
            help="Fitting function to use (rearranges parameters to make sense).",
        ),
        emittance_window_length=dict(
            help="Length of the moving average window. (# data points)",
            type=int,
            default=ROLLING_AVERAGE_WINDOW,
        ),
        emittance_outlier_limit=dict(
            help="Limit, i.e. cut from mean, on emittance outliers in meter.",
            type=float,
            default=OUTLIER_LIMIT,
        ),
        emittance_type=dict(
            type=str,
            default="average",
            choices=["fit_sigma", "average"],
            help="Which BSRT data to use (from database).",
        ),
        show=dict(action="store_true", help="Show plots.",),
        plot_styles=dict(
            type=str,
            nargs="+",
            default=["standard"],
            help="Which plotting styles to use, either from omc3 styles or default mpl.",
        ),
        manual_style=dict(
            type=DictAsString,
            default={},
            help="Additional style rcParameters which update the set of predefined ones.",
        ),
    )


@entrypoint(get_params(), strict=True)
def main(opt):
    LOG.debug("Starting Forced DA analysis.")
    _log_opt(opt)

    kick_dir, out_dir = _get_output_dir(opt.kick_directory, opt.output_directory)
    with suppress(PermissionError):
        save_config(out_dir, opt, __file__)

    # get data
    kick_df = _get_kick_df(kick_dir, opt.plane)
    intensity_df, emittance_df, emittance_bws_df = _get_dataframes(
        kick_df.index,
        opt.get_subdict(
            [
                "fill",
                "beam",
                "plane",
                "time_around_kicks",
                "emittance_tfs",
                "intensity_tfs",
                "show_wirescan_emittance",
                "timber_db",
                "pagestore_db",
                "emittance_window_length",
                "emittance_outlier_limit",
                "emittance_type",
                "normalized_emittance",
            ]
        ),
    )
    _check_all_times_in(kick_df.index, intensity_df.index[0], intensity_df.index[-1])

    # add data to kicks
    kick_df = _add_intensity_and_losses_to_kicks(
        kick_df, intensity_df, opt.intensity_time_before_kick, opt.intensity_time_after_kick
    )
    kick_df = _add_emittance_to_kicks(
        opt.plane, opt.energy, kick_df, emittance_df, opt.normalized_emittance
    )
    kick_df = _do_fit(opt.plane, kick_df, opt.fit)
    kick_df = _convert_to_sigmas(opt.plane, kick_df)

    # output
    _write_tfs(out_dir, opt.plane, kick_df, intensity_df, emittance_df, emittance_bws_df)

    # plotting
    figs = dict()
    register_matplotlib_converters()  # for datetime plotting
    style.set_style(opt.plot_styles, opt.manual_style)
    figs["emittance"] = _plot_emittances(
        out_dir, opt.beam, opt.plane, emittance_df, emittance_bws_df, kick_df.index
    )
    figs["intensity"] = _plot_intensity(out_dir, opt.beam, opt.plane, kick_df, intensity_df)
    for fit_type in ("exponential", "linear", "norm"):
        figs[f"da_fit_{fit_type}"] = _plot_da_fit(out_dir, opt.beam, opt.plane, kick_df, fit_type)

    if opt.show:
        plt.show()
    LOG.debug("Forced DA analysis finished.")
    return figs


# Helper ---


def _log_opt(opt: DotDict):
    """Show options in log."""
    LOG.info("Performing ForcedDA Analysis for:")
    if opt.fill is not None:
        LOG.info(f"  Fill: {opt.fill}")
    LOG.info(f"  Energy: {opt.energy} GeV")
    LOG.info(f"  Beam: {opt.beam}")
    LOG.info(f"  Plane: {opt.plane}")
    LOG.info(f"  Analysis Directory: '{opt.kick_directory}'")


def _write_tfs(
    out_dir: Path,
    plane: str,
    kick_df: DataFrame,
    intensity_df: DataFrame,
    emittance_df: DataFrame,
    emittance_bws_df: DataFrame,
):
    """Write out gathered data."""
    LOG.debug("Writing tfs files.")
    for df in (kick_df, intensity_df, emittance_df, emittance_bws_df):
        if df is not None:
            df.insert(0, TIME_COLUMN, [CERNDatetime(dt).cern_utc_string() for dt in df.index])
    try:
        tfs.write(out_dir / outfile_kick(plane), kick_df)
        tfs.write(out_dir / OUTFILE_INTENSITY, intensity_df)
        tfs.write(out_dir / outfile_emittance(plane), emittance_df)
        if emittance_bws_df is not None:
            tfs.write(out_dir / outfile_emittance_bws(plane), emittance_bws_df)
    except (FileNotFoundError, IOError):
        LOG.error(f"Cannot write into directory: {str(out_dir)} ")


def _check_all_times_in(series: Series, start: CERNDatetime, end: CERNDatetime):
    """Check if all times in series are between start and end."""
    if any(s for s in series if s < start or s > end):
        raise ValueError(
            "Some of the kick-times are outside of the fill times! "
            "Check if correct kick-file or fill number are used."
        )


def _convert_time_index(list_: list, path: Path = None) -> pd.Index:
    """Tries to convert time index to cerntime, first from datetime, then string, then timestamp."""
    for index_convert in (
        _datetime_to_cerntime_index,
        _string_to_cerntime_index,
        _timestamp_to_cerntime_index,
    ):
        with suppress(TypeError):
            return index_convert(list_)
    msg = f"Unrecognized format in column '{TIME_COLUMN}'"
    if path:
        msg += f" in '{str(path)}'"
    raise TypeError(msg)


def _string_to_cerntime_index(list_):
    return pd.Index((CERNDatetime.from_cern_utc_string(t) for t in list_), dtype=object)


def _timestamp_to_cerntime_index(list_):
    return pd.Index((CERNDatetime.from_timestamp(t) for t in list_), dtype=object)


def _datetime_to_cerntime_index(list_):
    return pd.Index((CERNDatetime(t) for t in list_), dtype=object)


def _drop_duplicate_indices(df):
    duplicate_mask = [True] + [
        df.index[idx] != df.index[idx - 1] for idx in range(1, len(df.index))
    ]
    return df.loc[duplicate_mask, :]


# TFS Data Loading -------------------------------------------------------------


def _get_dataframes(
    kick_times: pd.Index, opt: DotDict
) -> Tuple[TfsDataFrame, TfsDataFrame, TfsDataFrame]:
    """Gets the intensity and emittance dataframes from either input, files or (timber) database."""
    db = _get_db(opt)

    if opt.fill is not None:
        timespan_ts = _get_fill_times(db, opt.fill)
        timespan_dt = _convert_time_index(timespan_ts)
    else:
        td = pd.Timedelta(minutes=opt.time_around_kicks)
        timespan_dt = (kick_times.min() - td, kick_times.max() + td)
        timespan_ts = tuple(t.timestamp() for t in timespan_dt)

    if opt.intensity_tfs:
        intensity_df = _read_tfs(opt.intensity_tfs, timespan_dt)
    else:
        intensity_df = _get_bctrf_beam_intensity_from_timber(opt.beam, db, timespan_ts)

    if opt.emittance_tfs:
        emittance_df = _read_tfs(opt.emittance_tfs, timespan_dt)
    else:
        emittance_df = _get_bsrt_bunch_emittances_from_timber(
            opt.beam, opt.plane, db, timespan_ts, opt.emittance_type, opt.normalized_emittance
        )
    emittance_df = _filter_emittance_data(
        emittance_df, opt.plane, opt.emittance_window_length, opt.emittance_outlier_limit
    )

    if opt.show_wirescan_emittance is True:
        emittance_bws_df = _get_bws_emittances_from_timber(opt.beam, opt.plane, db, timespan_ts)
    elif opt.show_wirescan_emittance:
        emittance_bws_df = _read_tfs(opt.show_wirescan_emittance, timespan_dt)
    else:
        emittance_bws_df = None

    return intensity_df, emittance_df, emittance_bws_df


def _read_tfs(tfs_file_or_path, timespan):
    """Read previously gathered data (see :meth:`pylhc.forced_da_analysis._write_tfs`)."""
    try:
        tfs_df = tfs.read_tfs(tfs_file_or_path, index=TIME_COLUMN)
    except IOError:
        tfs_df = tfs_file_or_path  # hopefully

    tfs_df.index = _convert_time_index(tfs_df.index)

    return tfs_df.loc[slice(*timespan), :]


def _filter_emittance_data(df, planes, window_length, limit):
    """Cleans emittance data via outlier filter and moving average."""
    for plane in planes:
        LOG.debug(f"Filtering emittance data in plane {plane}.")
        col_nemittance = column_norm_emittance(plane)
        # col_err_nemittance = err_col(col_nemittance)
        col_mean = mean_col(col_nemittance)
        col_err_mean = err_col(col_mean)

        mav, std, mask = clean_outliers_moving_average(
            df[col_nemittance], length=window_length, limit=limit
        )
        df[col_mean] = mav
        df[col_err_mean] = std
        # if any(df[col_err_nemittance]):
        #     df[col_err_mean] = _rolling_errors(df[col_err_nemittance], ~mask, window_length)

    df = df.dropna(axis="index")
    if len(df.index) == 0:
        raise IndexError("Not enough emittance data extracted. Try to give a fill number.")

    df.headers[HEADER_BSRT_ROLLING_WINDOW] = window_length
    df.headers[HEADER_BSRT_OUTLIER_LIMIT] = limit
    df = _maybe_add_sum_for_planes(df, planes, column_norm_emittance)
    df = _maybe_add_sum_for_planes(
        df,
        planes,
        lambda p: mean_col(column_norm_emittance(p)),
        lambda p: err_col(mean_col(column_norm_emittance(p))),
    )
    return df


# Timber Data ------------------------------------------------------------------


def _get_db(opt):
    """Get the database either presaved or from timber."""
    db = None

    if opt.pagestore_db:
        db = opt.pagestore_db
        try:
            db_path = Path(db)
        except TypeError:
            pass
        else:
            LOG.debug(f"Loading database from file {str(db_path)}")
            db = PageStore(f"file:{str(db_path)}", str(db_path.with_suffix("")))
            if opt.fill is not None:
                raise EnvironmentError("'fill' can't be used with pagestore database.")
    else:
        LOG.debug(f" Trying to load database from timber.")
        try:
            db = pytimber.LoggingDB(source=opt["timber_db"])
        except AttributeError:
            LOG.debug(f" Loading from timber failed.")

    if not db:
        error_msg = ""
        if opt.fill is not None:
            error_msg += "'fill' is given, "
        if opt.emittance_tfs is None:
            error_msg += "'emittance_tfs' is not given, "
        if opt.intensity_tfs is None:
            error_msg += "'intensity_tfs' is not given, "
        if opt.show_wirescan_emittance is True:
            error_msg += "wirescan emittance is requested, "
        if len(error_msg):
            error_msg += (
                "but there is no database given and no access to timber databases. Aborting."
            )
            raise EnvironmentError(error_msg)
    return db


def _get_fill_times(db, fill):
    """Extract Fill times from database."""
    LOG.debug(f"Getting Timespan from fill {fill}")
    filldata = db.getLHCFillData(fill)
    return filldata["startTime"], filldata["endTime"]


def _get_bctrf_beam_intensity_from_timber(beam, db, timespan):
    LOG.debug(f"Getting beam intensity from bctfr for beam {beam}.")
    intensity_key = INTENSITY_KEY.format(beam=beam)
    LOG.debug(f"  Key: {intensity_key}")
    x, y = db.get(intensity_key, *timespan)[intensity_key]
    df = tfs.TfsDataFrame(
        data=y, index=_timestamp_to_cerntime_index(x), columns=[INTENSITY], dtype=float
    )

    df = _drop_duplicate_indices(df)
    LOG.debug(f"  Returning dataframe of shape {df.shape}")
    return df


def _get_bsrt_bunch_emittances_from_timber(beam, planes, db, timespan, key_type, nominal_emittance):
    dfs = {p: None for p in planes}
    for plane in planes:
        LOG.debug(f"Getting emittance from BSRT for beam {beam}  and plane {plane}.")
        bunch_emittance_key = bsrt_emittance_key(beam, plane, key_type)
        LOG.debug(f"  Key: {bunch_emittance_key}")
        col_nemittance = column_norm_emittance(plane)
        all_columns = [f(col_nemittance) for f in (lambda s: s, mean_col, err_col)] + [
            err_col(mean_col(col_nemittance))
        ]

        x, y = db.get(bunch_emittance_key, *timespan)[bunch_emittance_key]
        y_std = np.zeros_like(x)
        if key_type == "fit_sigma":
            # add all data with the same timestamp
            y_new = defaultdict(list)
            for x_elem, y_elem in zip(x, y):
                y_new[f"{x_elem:.3f}"] += y_elem.tolist()

            # get average and std per timestamp
            x = np.array([float(elem) for elem in y_new.keys()])
            y = np.array([np.average(elem) for elem in y_new.values()]) * nominal_emittance
            y_std = np.array([np.std(elem) for elem in y_new.values()]) * nominal_emittance
        elif key_type == "average":
            y *= BSRT_EMITTANCE_TO_METER
            y_std *= BSRT_EMITTANCE_TO_METER

        # remove entries with zero emittance as unphysical
        x, y, y_std = x[y != 0], y[y != 0], y_std[y != 0]

        df = tfs.TfsDataFrame(
            index=_timestamp_to_cerntime_index(x), columns=all_columns, dtype=float,
        )
        df[col_nemittance] = y
        df[err_col(col_nemittance)] = y_std

        dfs[plane] = df

    df = _merge_df_planes(dfs, planes)
    LOG.debug(f"  Returning dataframe of shape {df.shape}")
    return df


def _get_bws_emittances_from_timber(beam, planes, db, timespan):
    dfs = {p: None for p in planes}
    for plane in planes:
        LOG.debug(f"Getting emittance from BWS for beam {beam} and plane {plane}.")
        all_columns = [column_bws_norm_emittance(plane, d) for d in BWS_DIRECTIONS]
        df = None
        for direction in BWS_DIRECTIONS:
            emittance_key = bws_emittance_key(beam, plane, direction)
            LOG.debug(f"  Key: {emittance_key}")
            column_nemittance = column_bws_norm_emittance(plane, direction)

            x, y = db.get(emittance_key, *timespan)[emittance_key]
            if df is None:
                df = tfs.TfsDataFrame(
                    index=_timestamp_to_cerntime_index(x), columns=all_columns, dtype=float
                )
            df[column_nemittance] = y * BWS_EMITTANCE_TO_METER
            df[column_nemittance] = df[column_nemittance].apply(
                np.mean
            )  # BWS can give multiple values
            df[err_col(column_nemittance)] = df[column_nemittance].apply(
                np.std
            )  # BWS can give multiple values
            dfs[plane] = df

    df = _merge_df_planes(dfs, planes)
    for direction in BWS_DIRECTIONS:
        df = _maybe_add_sum_for_planes(
            df,
            planes,
            lambda p: column_bws_norm_emittance(p, direction),
            lambda p: err_col(column_bws_norm_emittance(p, direction)),
        )
    LOG.debug(f"  Returning dataframe of shape {df.shape}")
    return df


# Kick Data --------------------------------------------------------------------


def _get_kick_df(kick_dir, plane):
    def column_action_error(x):
        return err_col(column_action(x))

    try:
        df = _get_new_kick_file(kick_dir, plane)
    except FileNotFoundError:
        LOG.debug("Reading of kickfile failed. Looking for old kickfile.")
        df = _get_old_kick_file(kick_dir, plane)

    df = _maybe_add_sum_for_planes(df, plane, column_action, column_action_error)
    return df[[column_action(plane), column_action_error(plane)]]


def _get_old_kick_file(kick_dir, plane):
    """Kick files from ``Beta-Beat.src``."""
    path = kick_dir / "getkickac.out"
    LOG.debug(f"Reading kickfile '{str(path)}'.'")
    df = tfs.read(path)
    df = df.set_index(TIME_COLUMN)
    df.index = _convert_time_index(df.index, path)
    rename_dict = {}
    for p in plane:  # can be XY
        rename_dict.update(
            {
                f"2J{p}RES": column_action(p),
                f"2J{p}STDRES": err_col(column_action(p)),
                f"J{p}2": column_action(p),  # pre 2017
                f"J{p}STD": err_col(column_action(p)),  # pre 2017
            }
        )
    df = df.rename(rename_dict, axis="columns")
    renamed_cols = list(set(rename_dict.values()))
    df.loc[:, renamed_cols] = df.loc[:, renamed_cols] * 1e-6
    return df


def _get_new_kick_file(kick_dir, planes):
    """Kick files from ``omc3``."""
    dfs = {p: None for p in planes}
    for plane in planes:
        path = kick_dir / f"{KICKFILE}_{plane.lower()}{TFS_SUFFIX}"
        LOG.debug(f"Reading kickfile '{str(path)}'.'")
        df = tfs.read(path, index=TIME_COLUMN)
        df.index = pd.Index([CERNDatetime.from_cern_utc_string(t) for t in df.index], dtype=object)
        dfs[plane] = df
    return _merge_df_planes(dfs, planes)


def _get_output_dir(kick_directory, output_directory):
    kick_path = Path(kick_directory)
    if output_directory:
        output_path = Path(output_directory)
    else:
        output_path = kick_path / RESULTS_DIR
    try:
        output_path.mkdir(exist_ok=True)
    except PermissionError:
        LOG.warn(
            f"You have no writing permission in '{str(output_path)}', "
            f"output data might not be created."
        )
    LOG.info(f"All output will be written to {str(output_path)}")
    return kick_path, output_path


# Intensity at Kicks -----------------------------------------------------------


def _add_intensity_and_losses_to_kicks(kick_df, intensity_df, time_before, time_after):
    LOG.debug("Calculating intensity and losses for the kicks.")
    col_list = [INTENSITY_BEFORE, INTENSITY_AFTER, INTENSITY_LOSSES]
    new_columns = [col for col in col_list + [err_col(c) for c in col_list]]
    kick_df = kick_df.reindex(columns=kick_df.columns.tolist() + new_columns)
    kick_df = _get_intensities_around_kicks(kick_df, intensity_df, time_before, time_after)
    kick_df = _calculate_intensity_losses_at_kicks(kick_df)
    return kick_df


def _get_intensities_around_kicks(kick_df, intensity_df, time_before, time_after):
    LOG.debug("Calculating beam intensity before and after kicks.")
    # input signs and order does not matter
    time_before = sorted(-np.abs(t) for t in time_before)
    time_after = sorted(np.abs(t) for t in time_after)

    kick_df.headers[HEADER_TIME_BEFORE] = str(time_before)
    kick_df.headers[HEADER_TIME_AFTER] = str(time_after)
    for i, time in enumerate(kick_df.index):
        # calculate intensity before and after kicks (with error)
        for column, time_delta in ((INTENSITY_BEFORE, time_before), (INTENSITY_AFTER, time_after)):
            t_from, t_to = (
                time + pd.Timedelta(seconds=time_delta[0]),
                time + pd.Timedelta(seconds=time_delta[1]),
            )
            data = intensity_df.loc[
                t_from:t_to, INTENSITY
            ]  # awesome pandas can handle time intervals!
            kick_df.loc[time, [column, err_col(column)]] = data.mean(), data.std()
    return kick_df


def _calculate_intensity_losses_at_kicks(kick_df):
    LOG.debug("Calculating intensity losses.")
    # absolute losses
    kick_df[INTENSITY_LOSSES] = kick_df[INTENSITY_BEFORE] - kick_df[INTENSITY_AFTER]
    kick_df[err_col(INTENSITY_LOSSES)] = np.sqrt(
        np.square(kick_df[err_col(INTENSITY_BEFORE)]) + np.square(kick_df[err_col(INTENSITY_AFTER)])
    )

    # relative losses, error from error-propagation formular for losses / I_before = 1 - I_after / I_before
    kick_df[rel_col(INTENSITY_LOSSES)] = kick_df[INTENSITY_LOSSES] / kick_df[INTENSITY_BEFORE]
    kick_df[rel_col(err_col(INTENSITY_LOSSES))] = np.sqrt(
        np.square(kick_df[INTENSITY_AFTER] / kick_df[INTENSITY_BEFORE])
        * (
            np.square(kick_df[err_col(INTENSITY_AFTER)] / kick_df[INTENSITY_AFTER])
            + np.square(kick_df[err_col(INTENSITY_BEFORE)] / kick_df[INTENSITY_BEFORE])
        )
    )
    return kick_df


# Emittance at Kicks -----------------------------------------------------------


def _add_emittance_to_kicks(plane, energy, kick_df, emittance_df, nominal):
    LOG.debug("Retrieving normalized emittance at the kicks.")
    kick_df.headers[HEADER_ENERGY] = energy
    kick_df.headers[HEADER_BSRT_ROLLING_WINDOW] = ROLLING_AVERAGE_WINDOW
    col_nemittance = column_norm_emittance(plane)
    cols_emitt = [mean_col(col_nemittance), err_col(mean_col(col_nemittance))]
    cols_kick = [col_nemittance, err_col(col_nemittance)]

    kick_df = kick_df.reindex(columns=kick_df.columns.tolist() + cols_kick)
    idx_emitt = [emittance_df.columns.get_loc(c) for c in cols_emitt]
    for time in kick_df.index:
        idx_kick = emittance_df.index.get_loc(time, method="nearest")
        kick_df.loc[time, cols_kick] = emittance_df.iloc[idx_kick, idx_emitt].values

    # add de-normalized emittance
    normalization = get_proton_gamma(energy) * get_proton_beta(
        energy
    )  # norm emittance to emittance
    col_emittance = column_emittance(plane)

    kick_df.headers[header_norm_nominal_emittance(plane)] = nominal
    kick_df.headers[header_nominal_emittance(plane)] = nominal / normalization
    kick_df[col_emittance] = kick_df[col_nemittance] / normalization
    kick_df[err_col(col_emittance)] = kick_df[err_col(col_nemittance)] / normalization
    return kick_df


# Forced DA Fitting ------------------------------------------------------------


def fun_exp_decay(p, x):  # fit and plot
    """sp = DA_J, x[0] = action (2J res), x[1] = emittance"""
    return np.exp(-(p - (0.5 * x[0])) / x[1])


def fun_exp_sigma(p, x):  # only used for plotting
    """p = DA_sigma, x = action (J_sigma)"""
    return np.exp(-0.5 * (p ** 2 - x ** 2))


def fun_linear(p, x):  # fit and plot
    """p = DA_J, x = action (2J res)"""
    return x * 0.5 - p


def swap_fun_parameters(fun):
    """Parameter swapped for Curvefit."""
    return lambda x, p: fun(p, x)


def _do_fit(plane, kick_df, fit_type):
    LOG.debug("Fitting forced da to exponential. ")
    action, emittance, rel_losses = _get_fit_data(kick_df, plane)
    init_guess = [INITIAL_DA_FIT * kick_df.headers[header_nominal_emittance(plane)]]

    get_fit_param = {"linear": _linear_fit_parameters, "exponential": _exponential_fit_parameters}[
        fit_type
    ]

    fit_fun, x, y, sx, sy = get_fit_param(action, emittance, rel_losses)

    # do prelim fit
    init_fit, _ = _fit_curve(swap_fun_parameters(fit_fun), x, y, init_guess)

    # do odr
    odr = _fit_odr(fit_fun, x, y, sx, sy, init_fit)

    # add DA to kick
    da = odr.beta[0], odr.sd_beta[0]
    kick_df.headers[header_da(plane)], kick_df.headers[header_da_error(plane)] = da
    LOG.info(f"Forced DA (wrt. J) in {plane} [m]: {da[0]} ± {da[1]}")

    return kick_df


def _get_fit_data(kick_df, plane):
    """Extracts necessary data from ``kick-df``. Returns tri-tuple of tuples (data, std)."""
    col_action = column_action(plane)
    col_emittance = column_emittance(plane)
    col_losses = rel_col(INTENSITY_LOSSES)

    # get data
    action = kick_df[col_action], _no_nonzero_errors(kick_df[err_col(col_action)])
    emittance = kick_df[col_emittance], _no_nonzero_errors(kick_df[err_col(col_emittance)])
    rel_losses = kick_df[col_losses], _no_nonzero_errors(kick_df[err_col(col_losses)])
    return action, emittance, rel_losses


def _exponential_fit_parameters(action, emittance, rel_losses):
    """Returns exponential fit function and parameters. All inputs are tuples of (data, std)."""
    x = action[0], emittance[0]
    y = rel_losses[0]

    sx = [action[1], emittance[1]]
    sy = rel_losses[1]
    return fun_exp_decay, x, y, sx, sy


def _linear_fit_parameters(action, emittance, rel_losses):
    """
    Returns linear fit function and parameters. All inputs are tuples of (data, std)."""
    log_losses = np.log(rel_losses[0])
    x = action[0]
    y = emittance[0] * log_losses

    sx = action[1]
    sy = np.sqrt(
        (log_losses * emittance[1]) ** 2 + ((emittance[0] * rel_losses[1]) / rel_losses[0]) ** 2
    )
    return fun_linear, x, y, sx, sy


def _fit_curve(fun, x, y, init):
    """Initial curve fit, without errors."""
    fit, cov = scipy.optimize.curve_fit(fun, x, y, p0=init, maxfev=MAX_CURVEFIT_FEV)
    LOG.info(f"Initial DA fit: {fit} with cov {cov}")
    return fit, np.sqrt(np.diag(cov))


def _fit_odr(fun, x, y, sx, sy, init):
    """ODR Fit (includes errors)."""
    # fill zero errors with the minimum error - otherwise fit will not work
    fit_model_sigma = scipy.odr.Model(fun)
    data_model_sigma = scipy.odr.RealData(x=x, y=y, sx=sx, sy=sy,)
    da_odr = scipy.odr.ODR(data_model_sigma, fit_model_sigma, beta0=init)
    # da_odr.set_job(fit_type=2)
    odr_output = da_odr.run()
    logging_tools.odr_pprint(LOG.info, odr_output)
    return odr_output


def _no_nonzero_errors(series):
    """Removes all zero-erros and replaces them with minimum errors in set."""
    series = series.copy()
    nonzero = series[series != 0]
    if len(nonzero) == 0:
        raise ValueError("All errors are exact zero. Can't do ODR fit.")
    series[series == 0] = np.abs(series[series != 0]).min()
    return series


def _convert_to_sigmas(plane, kick_df):
    """Converts the DA and the Action into Sigma-Units."""
    LOG.debug("Calculating action and da in sigmas.")
    nominal_emittance = kick_df.headers[header_nominal_emittance(plane)]
    emittance = kick_df[column_emittance(plane)]
    emittance_mean, emittance_std = emittance.mean(), emittance.std()
    emittance_sign, emittance_sign_std = significant_digits(
        emittance_mean * 1e12, emittance_std * 1e12
    )
    LOG.info(
        f"Measured Emittance {emittance_sign} ± {emittance_sign_std} pm"
        f" (Nominal {nominal_emittance*1e12: .2f} pm)"
    )

    # DA (in units of J) to DA_sigma
    da, da_err = kick_df.headers[header_da(plane)], kick_df.headers[header_da_error(plane)]
    da_sigma, da_sigma_err = (
        np.sqrt(2 * da / emittance_mean),
        da_err / np.sqrt(2 * da * emittance_mean),
    )
    kick_df.headers[header_da(plane, unit="sigma")] = da_sigma
    kick_df.headers[header_da_error(plane, unit="sigma")] = da_sigma_err
    LOG.info(f"Forced DA {plane} in N-sigma: {da_sigma} ± {da_sigma_err}")

    # Action (in units of 2J) to J_sigma
    col_action = column_action(plane)
    # kick_df[sigma_col(col_action)] = np.sqrt(kick_df[col_action] / nominal_emittance)
    # kick_df[err_col(sigma_col(col_action))] = (
    #         0.5 * kick_df[err_col(col_action)] / np.sqrt(kick_df[col_action] * nominal_emittance)
    # )
    kick_df[sigma_col(col_action)] = np.sqrt(kick_df[col_action] / emittance)
    kick_df[err_col(sigma_col(col_action))] = (
        0.5 * kick_df[err_col(col_action)] / np.sqrt(kick_df[col_action] * emittance)
    )
    return kick_df


# Plotting ---------------------------------------------------------------------


def _plot_intensity(directory, beam, plane, kick_df, intensity_df):
    """
    Plots beam intensity. For losses, the absolute values are used and then normalized
    to the Intensity before the kicks, to get the percentage relative to that (global) value.
    """
    LOG.debug("Plotting beam intensity")
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(16.80, 7.68))

    x_span = (kick_df.index.max() - kick_df.index.min()).seconds * np.array(
        [0.03, 0.09]
    )  # defines x-limits

    # convert to % relative to before first kick
    idx_before = intensity_df.index.get_loc(
        kick_df.index.min() - pd.Timedelta(seconds=x_span[0]), method="ffill"
    )
    idx_intensity = intensity_df.columns.get_loc(INTENSITY)  # for iloc
    intensity_start = intensity_df.iloc[idx_before, idx_intensity]
    norm = intensity_start / 100.0

    # plot intensity
    ax.plot(
        _date2num(intensity_df.index),
        intensity_df[INTENSITY] / norm,
        marker=".",
        markersize=mpl.rcParams["lines.markersize"] * 0.5,
        fillstyle="full",
        color=colors.get_mpl_color(0),
        label=f"Intensity",
    )

    # plot losses per kick
    normalized_intensity = kick_df.loc[:, [INTENSITY_BEFORE, INTENSITY_AFTER]] / norm
    normalized_intensity_error = (
        kick_df.loc[:, [err_col(INTENSITY_BEFORE), err_col(INTENSITY_AFTER)]] / norm
    )
    normalized_losses = kick_df.loc[:, [INTENSITY_LOSSES, err_col(INTENSITY_LOSSES)]] / norm
    normalized_losses_kick = (
        kick_df.loc[:, [rel_col(INTENSITY_LOSSES), err_col(rel_col(INTENSITY_LOSSES))]] * 100
    )

    for idx, kick in enumerate(kick_df.index):
        ax.errorbar(
            [_date2num(kick)] * 2,
            normalized_intensity.loc[kick, :],
            yerr=normalized_intensity_error.loc[kick, :],
            color=colors.get_mpl_color(1),
            marker=".",
            linestyle="-",
            label="__nolegend__" if idx > 0 else "Losses",
        )

        ax.text(
            _date2num(kick),
            0.5 * sum(normalized_intensity.loc[kick, :]),
            "  -{:.1f}$\pm${:.1f} %\n".format(*normalized_losses.loc[kick, :])
            + " (-{:.1f}$\pm${:.1f} %)".format(*normalized_losses_kick.loc[kick, :]),
            va="bottom",
            color=colors.get_mpl_color(1),
            fontdict=dict(fontsize=mpl.rcParams["font.size"] * 0.8),
        )

    _plot_kicks_and_scale_x(ax, kick_df.index, pad=x_span)
    ylim = [normalized_intensity.min().min(), normalized_intensity.max().max()]
    ypad = 0.1 * (ylim[1] - ylim[0])
    ax.set_ylim([ylim[0] - ypad, ylim[1] + ypad])
    ax.set_ylabel(r"Beam Intensity [%]")
    annotations.make_top_legend(ax, ncol=3)
    plt.tight_layout()
    annotations.set_name(f"Intensity Beam {beam}, Plane {plane}", fig)
    annotations.set_annotation(
        f"Intensity at 100%: {intensity_start*1e-10:.3f}" "$\;\cdot\;10^{{10}}$ charges",
        ax=ax,
        position="left",
    )
    _save_fig(directory, plane, fig, "intensity")
    return fig


def _plot_emittances(directory, beam, plane, emittance_df, emittance_bws_df, kick_times):
    LOG.debug("Plotting normalized emittances")
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10.24, 7.68))

    col_norm_emittance = column_norm_emittance(plane)

    bsrt_color = colors.get_mpl_color(0)
    bws_color = colors.get_mpl_color(1)

    ax.errorbar(
        _date2num(emittance_df.index),
        emittance_df[col_norm_emittance] * 1e6,  # Actual BSRT measurement
        yerr=emittance_df[err_col(col_norm_emittance)] * 1e6,
        color=bsrt_color,
        marker="o",
        markeredgewidth=2,
        linestyle="None",
        label=f"From BSRT",
    )

    ax.errorbar(
        _date2num(emittance_df.index),
        emittance_df[mean_col(col_norm_emittance)] * 1e6,
        yerr=emittance_df[err_col(mean_col(col_norm_emittance))] * 1e6,
        color=colors.change_color_brightness(bsrt_color, 0.7),
        marker="",
        label=f"Moving Average (window = {ROLLING_AVERAGE_WINDOW})",
    )

    if emittance_bws_df is not None and len(emittance_bws_df.index):
        for d in BWS_DIRECTIONS:
            label = "__nolegend__" if d == BWS_DIRECTIONS[1] else f"From BWS"
            color = (
                bws_color
                if d == BWS_DIRECTIONS[1]
                else colors.change_color_brightness(bws_color, 0.5)
            )
            col_bws_nemittance = column_bws_norm_emittance(plane, d)
            ax.errorbar(
                _date2num(emittance_bws_df.index),
                emittance_bws_df[col_bws_nemittance] * 1e6,
                yerr=emittance_bws_df[err_col(col_bws_nemittance)] * 1e6,
                linestyle="None",
                marker="o",
                color=color,
                label=label,
                markersize=mpl.rcParams["lines.markersize"] * 1.5,
            )

    _plot_kicks_and_scale_x(ax, kick_times)
    ax.set_ylabel(r"$\epsilon_{n}$ $[\mu m]$")
    annotations.make_top_legend(ax, ncol=2)
    plt.tight_layout()
    annotations.set_name(f"Emittance Beam {beam}, Plane {plane}", fig)
    _save_fig(directory, plane, fig, "emittance")
    return fig


def _plot_da_fit(directory, beam, plane, k_df, fit_type):
    """
    Plot the Forced Dynamic Aperture fit.
    (I do not like the complexity of this function. jdilly).
    """
    LOG.debug(f"Plotting Dynamic Aperture Fit for {fit_type}")
    col_action = column_action(plane)
    col_action_sigma = sigma_col(col_action)
    col_emittance = column_emittance(plane)
    col_intensity = rel_col(INTENSITY_LOSSES)

    kick_df = k_df.copy()
    kick_df = kick_df.sort_values(by=col_action)

    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10.24, 7.68))

    # Plot Measurement Data
    intensity = kick_df[col_intensity]
    intensity_err = kick_df[err_col(col_intensity)]
    if fit_type == "linear":
        intensity_err = np.abs(1 / intensity) * intensity_err
        intensity = np.log(intensity)
    else:
        intensity *= 100
        intensity_err *= 100

    if fit_type == "norm":
        action = kick_df[col_action_sigma]
        action_err = kick_df[err_col(col_action_sigma)]
        action_x = action
        action_xerr = action_err
    else:
        action = kick_df[col_action]
        action_err = kick_df[err_col(col_action)]
        action_x = action * 1e6
        action_xerr = action_err * 1e6

    ax.errorbar(
        action_x,
        intensity,
        xerr=action_xerr,
        yerr=intensity_err,
        marker=".",
        color=colors.get_mpl_color(0),
        label=f"Kicks",
    )

    # Plot Fit
    emittance = kick_df[col_emittance]
    da, da_err = kick_df.headers[header_da(plane)], kick_df.headers[header_da_error(plane)]
    da_mu, da_err_mu = significant_digits(da * 1e6, da_err * 1e6)
    da_label = f"Fit: DA$_J$= ${da_mu} \pm {da_err_mu} \mu m$"

    if fit_type == "linear":
        fit_fun = fun_linear
        fit_data = action
        multiplier = 1 / emittance  # DA-J/emittance = -ln(I/Io)
    elif fit_type == "exponential":
        fit_fun = fun_exp_decay
        fit_data = (action, emittance)
        multiplier = 100  # for percentages
    elif fit_type == "norm":
        da, da_err = (
            kick_df.headers[header_da(plane, unit="sigma")],
            kick_df.headers[header_da_error(plane, unit="sigma")],
        )
        da_round, da_err_round = significant_digits(da, da_err)
        da_label = f"Fit: DA= ${da_round} \pm {da_err_round} N_{{\sigma}}$"
        fit_fun = fun_exp_sigma
        fit_data = action
        multiplier = 100  # for percentages

    fit_mean = fit_fun(da, fit_data) * multiplier
    fit_min = fit_fun(da - da_err, fit_data) * multiplier
    fit_max = fit_fun(da + da_err, fit_data) * multiplier

    color = colors.get_mpl_color(1)
    ax.fill_between(action_x, fit_min, fit_max, facecolor=mcolors.to_rgba(color, 0.3))

    ax.plot(action_x, fit_mean, ls="--", c=color, label=da_label)

    # extend fit to 100% losses
    color_ext = "#7f7f7f"
    action_max = action.max()
    emittance_at_max = emittance[action == action_max][0]

    if fit_type in ["linear", "exponential"]:
        da_x = da * 2 * 1e6
        da_string = "2DA$_J$"
    elif fit_type == "norm":
        da_x = da
        da_string = "DA$_\sigma$"

    if action_max < da:
        if fit_type in ["linear", "exponential"]:
            action_ext = np.linspace(action_max, 2 * da, 10)
            action_x_ext = action_ext * 1e6
            if fit_type == "exponential":
                fit_data_ext = (action_ext, emittance_at_max)
            elif fit_type == "linear":
                fit_data_ext = action_ext
                multiplier = 1 / emittance_at_max
        else:
            action_ext = np.linspace(action_max, da, 10)
            action_x_ext = action_ext
            fit_data_ext = action_ext

        fit_ext = fit_fun(da, fit_data_ext) * multiplier
        ax.plot(
            action_x_ext,
            fit_ext,
            ls="--",
            color=mcolors.to_rgba(color_ext, 0.3),
            label="__nolegend__",
        )
    # DA Marker
    ax.axvline(da_x, ls="--", color=color_ext, marker="", label="__nolegend__")
    trans = mtrans.blended_transform_factory(ax.transData, ax.transAxes)  # x is data, y is axes
    ax.text(
        x=da_x,
        y=1.0,
        s=da_string,
        va="bottom",
        ha="center",
        zorder=-1,
        color=color_ext,
        transform=trans,
    )

    # Format figure
    if fit_type == "norm":
        nominal_emittance = kick_df.headers[header_nominal_emittance(plane)]
        emittance_mean, emittance_std = emittance.mean(), emittance.std()
        emittance_sign, emittance_sign_std = significant_digits(
            emittance_mean * 1e12, emittance_std * 1e12
        )
        ax.text(
            x=0,
            y=1.00,
            s=(
                f"$\epsilon_{{mean}}$ = {emittance_sign} $\pm$ {emittance_sign_std} pm "
                f"($\epsilon_{{nominal}}$ = {nominal_emittance*1e12: .2f} pm)"
            ),
            transform=ax.transAxes,
            va="bottom",
            ha="left",
        )
        ax.set_xlabel(
            f"$N_{{\sigma}} = \sqrt{{2J_{{{plane if len(plane) == 1 else ''}}}/\epsilon}}$"
        )
    else:
        ax.set_xlabel(f"$2J_{{{plane if len(plane) == 1 else ''}}} \; [\mu m]$")

    if fit_type == "linear":
        ax.set_ylabel(r"ln($I/I_0$)")
    else:
        ax.set_ylabel(r"Beam Losses [%]")
        ax.set_ylim([0, intensity.max() * (1 + YPAD)])
    ax.set_xlim([0, None])
    annotations.make_top_legend(ax, ncol=3)
    plt.tight_layout()
    annotations.set_name(
        f"DA {'' if fit_type == 'norm' else 'J'} {fit_type} Fit {beam}, Plane {plane}", fig
    )
    _save_fig(directory, plane, fig, f"dafit_{fit_type}")
    return fig


def _get_fit_plot_data(da, da_err, data, fit_type):
    fit_fun = {"exponential": fun_exp_decay, "linear": fun_linear}[fit_type]
    multiplier = 100  # for percentages
    if fit_type == "linear":
        multiplier = 1 / data[1]  # DA-J/emittance = -ln(I/Io)
        data = data[0]
    fit_mean = fit_fun(da, data) * multiplier
    fit_min = fit_fun(da - da_err, data) * multiplier
    fit_max = fit_fun(da + da_err, data) * multiplier
    return fit_mean, fit_min, fit_max


# Helper ---


def _plot_kicks_and_scale_x(ax, kick_times, pad=20):
    lines.plot_vertical_lines_fast(
        ax, kick_times, color="grey", linestyle="--", alpha=0.8, marker="", label="Kicks"
    )

    first_kick, last_kick = kick_times.min(), kick_times.max()
    try:
        time_delta = [pd.Timedelta(seconds=pad[i]) for i in range(2)]
    except TypeError:
        time_delta = [pd.Timedelta(seconds=pad) for _ in range(2)]

    # ax.set_xlim([(first_kick - time_delta[0]).timestamp, last_kick + time_delta[1]])  # worked in the past
    ax.set_xlim([_date2num(first_kick - time_delta[0]), _date2num(last_kick + time_delta[1])])
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M:%S"))
    ax.set_xlabel("Time")
    annotations.set_annotation(f"Date: {first_kick.strftime('%Y-%m-%d')}", ax, "left")


def _merge_df_planes(df_dict, planes):
    """In case planes == 'XY' merge the ``df_dict`` into one dataframe.."""
    if len(planes) == 1:
        return df_dict[planes]
    return pd.merge(*df_dict.values(), how="inner", left_index=True, right_index=True)


def _maybe_add_sum_for_planes(df, planes, col_fun, col_err_fun=None):
    """In case planes == 'XY' add the two plane columns and their errors."""
    if len(planes) > 1:
        if col_err_fun is not None:
            cols = lambda p: [col_fun(p), col_err_fun(p)]
            x_cols, y_cols = [cols(p) for p in planes]
            df = df.reindex(columns=df.columns.to_list() + cols(planes))
            df[cols(planes)] = np.array(
                toolbox.df_sum_with_err(
                    df, a_col=x_cols[0], b_col=y_cols[0], a_err_col=x_cols[1], b_err_col=y_cols[1]
                )
            ).T
        else:
            x_col, y_col = [col_fun(p) for p in planes]
            df[col_fun(planes)] = toolbox.df_sum(df, a_col=x_col, b_col=y_col)
    return df


def _date2num(times):
    """
    Convert CERNDatetime to mpl-number (days).

    Converts input times to plain-datetime first as date2num causes infinite loop with
    CernDateTimes in **Python 3.8**.
    """
    try:
        times = [cdt.datetime for cdt in times]
    except AttributeError:
        pass  # probably datetime already
    except TypeError:
        try:  # not iterable
            times = times.datetime
        except AttributeError:
            pass  # probably datetime already
    return mdates.date2num(times)


def _save_fig(directory, plane, fig, ptype):
    try:
        for ftype in PLOT_FILETYPES:
            path = os.path.join(directory, outfile_plot(ptype, plane, ftype))
            LOG.debug(f"Saving Figure to {path}")
            fig.savefig(path)
    except IOError:
        LOG.error(f"Couldn't create output files for {ptype} plots.")


# Script Mode ------------------------------------------------------------------


if __name__ == "__main__":
    main()



"""
Kickgroups
----------

Functions to list KickGroups and show their Kicks.

.. code-block:: none

    usage: kickgroups.py [-h] {kickgroups,kickgroup_info} ...

    KickGroups Functions

    optional arguments:
      -h, --help            show this help message and exit

    Functionality:
      {kickgroups,kickgroup_info}
        kickgroups          List all KickGroups
        kickgroup_info      Show the info of a given KickGroup


Function ``kickgroups``:

.. code-block:: none

    usage: kickgroups.py kickgroups [-h] [--root ROOT]
                                    [--sort {TIMESTAMP,KICKGROUP}]

    KickGroups

    optional arguments:
      -h, --help            show this help message and exit
      --root ROOT           KickGroups Root-Directory
      --sort {TIMESTAMP,KICKGROUP}
                            Sort KickGroups


Function ``kickgroup_info``:

.. code-block:: none

    usage: kickgroups.py kickgroup_info [-h] [--root ROOT] name

    KickGroup Info

    positional arguments:
      name         KickGroup name

    optional arguments:
      -h, --help   show this help message and exit
      --root ROOT  KickGroups Root-Directory

"""
import argparse
import json

from datetime import datetime
from pathlib import Path
from typing import List, Union

import numpy as np
import pandas as pd

from dateutil import tz
from omc3.utils import logging_tools
from pandas import DataFrame
from tfs import TfsDataFrame

# fmt: off
from pylhc.constants.kickgroups import (
    AMPX,
    AMPY,
    AMPZ,
    BEAM,
    BEAMPROCESS,
    BUNCH,
    COLUMNS_TO_HEADERS,
    DRIVEN_TUNEX,
    DRIVEN_TUNEY,
    DRIVEN_TUNEZ,
    KICK_COLUMNS,
    KICK_GROUP_COLUMNS,
    KICKGROUP,
    KICKGROUPS_ROOT,
    LOCALTIME,
    OPTICS,
    OPTICS_URI,
    SDDS,
    TIMESTAMP,
    TUNEX,
    TUNEY,
    TURNS,
    UTCTIME,
)

# fmt: on

LOG = logging_tools.get_logger(__name__)

# List Kickgroups --------------------------------------------------------------


def list_available_kickgroups(by: str = TIMESTAMP, root: Union[Path, str] = KICKGROUPS_ROOT, printout: bool = True) -> DataFrame:
    """
    List all available KickGroups in `root` with optional sorting..

    Args:
        by (str): Column to sort the KickGroups by. Should be either the
            ``TIMESTAMP`` or ``KICKGROUP`` variable.
        root (Path): Alternative `~pathlib.Path` to the KickGroup folder. (Defaults
            to the ``NFS`` path of our kickgroups).
        printout (bool): whether to print out the dataframe, defaults to `True`.

    Returns:
        A `~pandas.DataFrame` with the KickGroups loaded, sorted by the provided
        *by* parameter.
    """
    LOG.debug(f"Listing KickGroups in '{Path(root).absolute()}'")
    kickgroup_paths = get_folder_json_files(root)
    df_info = DataFrame(index=range(len(kickgroup_paths)), columns=KICK_GROUP_COLUMNS)
    for idx, kick_group in enumerate(kickgroup_paths):
        LOG.debug(f"Loading kickgroup info from '{kick_group.absolute()}'")
        data = _load_json(kick_group)
        df_info.loc[idx, KICKGROUP] = data["groupName"]
        df_info.loc[idx, TIMESTAMP] = data["groupCreationTime"]
        df_info.loc[idx, UTCTIME] = _ts_to_datetime(df_info.loc[idx, TIMESTAMP])
        df_info.loc[idx, LOCALTIME] = _utc_to_local(df_info.loc[idx, UTCTIME])
    df_info = df_info.sort_values(by=by).set_index(TIMESTAMP)

    if printout:
        LOG.debug(f"Here is information about the loaded KickGroups")
        print(df_info.to_string(index=False, formatters=_time_formatters(), justify="center"))

    return df_info


def get_folder_json_files(root: Union[Path, str] = KICKGROUPS_ROOT) -> List[Path]:
    """Returns a list of all **.json** files in the folder.

    Args:
        root (Union[Path, str])): the path to the folder. (Defaults
            to the ``NFS`` path of our kickgroups).

    Returns:
        A `list` of `~pathlib.Path` objects to all **json** files within the provided
        *root* parameter.
    """
    LOG.debug(f"Globing for JSON files in {Path(root).absolute()}''")
    return list(Path(root).glob("*.json"))


# Kickgroup Info ---------------------------------------------------------------


def kickgroup_info(kick_group: str, root: Union[Path, str] = KICKGROUPS_ROOT, printout: bool = True) -> TfsDataFrame:
    """
    Gather all important info about the KickGroup into a `~tfs.TfsDataFrame` and print it.

    Args:
        kick_group (str): the KickGroup name, corresponds to the kickgroup file name without
            the ``.json`` extension.
        root (Path): Alternative `~pathlib.Path` to the KickGroup folder. (Defaults
            to the ``NFS`` path of our kickgroups).
        printout (bool): whether to print out the dataframe, defaults to `True`.

    Returns:
        A `~tfs.TfsDataFrame` with the KickGroup information loaded.
    """
    LOG.debug(f"Loading info from all KickFiles in KickGroup '{kick_group}'")
    kick_group_data = _load_json(Path(root) / f"{kick_group}.json")
    kicks_files = kick_group_data["jsonFiles"]
    df_info = TfsDataFrame(index=range(len(kicks_files)), columns=KICK_COLUMNS, headers={KICKGROUP: kick_group})

    for idx, kf in enumerate(kicks_files):
        df_info.loc[idx, :] = load_kickfile(kf)

    for column in COLUMNS_TO_HEADERS:
        df_info.headers[column] = df_info[column][0]

    if printout:
        _print_kickgroup_info(df_info)

    return df_info


def load_kickfile(kickfile: Union[Path, str]) -> pd.Series:
    """
    Load the important data from a **json** kickfile into a `~pandas.Series`.

    Args:
        kickfile (Union[Path, str]): the path to the kickfile to load data from.

    Returns:
        A `~pandas.Series` with the relevant information loaded from the provided
        *kickfile*. The various entries in the Series are defined in `pylhc.constants.kickgroups`
        as ``KICK_COLUMNS``.
    """
    LOG.debug(f"Loading kick information from Kickfile at '{Path(kickfile).absolute()}'")
    kick = _load_json(kickfile)
    data = pd.Series(index=KICK_COLUMNS, dtype=object)
    data[LOCALTIME] = _jsontime_to_datetime(kick["acquisitionTime"])
    data[UTCTIME] = _local_to_utc(data[LOCALTIME])
    data[SDDS] = kick["sddsFile"]
    data[BEAM] = kick["measurementEnvironment"]["lhcBeam"]["beamName"]
    data[BEAMPROCESS] = kick["measurementEnvironment"]["environmentContext"]["name"]
    data[TURNS] = kick["acqSettings"]["capturedTurns"]
    data[BUNCH] = kick["acqSettings"]["bunchSelection"]
    data[OPTICS] = kick["measurementEnvironment"]["opticsModel"]["opticName"]
    data[OPTICS_URI] = kick["measurementEnvironment"]["opticsModel"]["opticModelURI"]

    three_d = "3D" in kick["excitationSettings"][0]["type"]

    if three_d:
        LOG.debug("Kick is 3D Excitation, loading longitudinal kick settings")
        data[TUNEX] = kick["excitationSettings"][0]["acDipoleSettings"][0]["measuredTune"]
        data[TUNEY] = kick["excitationSettings"][0]["acDipoleSettings"][1]["measuredTune"]
        data[DRIVEN_TUNEX] = data[TUNEX] + kick["excitationSettings"][0]["acDipoleSettings"][0]["deltaTuneStart"]
        data[DRIVEN_TUNEY] = data[TUNEY] + kick["excitationSettings"][0]["acDipoleSettings"][1]["deltaTuneStart"]
        data[DRIVEN_TUNEZ] = kick["excitationData"][0]["rfdata"]["excitationFrequency"]
        data[AMPX] = kick["excitationSettings"][0]["acDipoleSettings"][0]["amplitude"]
        data[AMPY] = kick["excitationSettings"][0]["acDipoleSettings"][1]["amplitude"]
        data[AMPZ] = kick["excitationSettings"][0]["longitudinalRfSettings"]["excitationAmplitude"]
    else:
        LOG.debug(f"Kick is 2D Excitation, longitudinal settings will be set as NaNs")
        data[TUNEX] = kick["excitationSettings"][0]["measuredTune"]
        data[TUNEY] = kick["excitationSettings"][1]["measuredTune"]
        data[DRIVEN_TUNEX] = data[TUNEX] + kick["excitationSettings"][0]["deltaTuneStart"]
        data[DRIVEN_TUNEY] = data[TUNEY] + kick["excitationSettings"][1]["deltaTuneStart"]
        data[DRIVEN_TUNEZ] = np.NaN
        data[AMPX] = kick["excitationSettings"][0]["amplitude"]
        data[AMPY] = kick["excitationSettings"][1]["amplitude"]
        data[AMPZ] = np.NaN

    return data


def _print_kickgroup_info(kicks_info: TfsDataFrame):
    """Actually print the full info about the kickgroup."""
    for header, value in kicks_info.headers.items():
        print(f"{header}: {value}")
    print()
    print(
        kicks_info.drop(columns=COLUMNS_TO_HEADERS).to_string(
            index=False, na_rep=" - ", justify="center", formatters=_time_formatters()
        )
    )


# Helper -----------------------------------------------------------------------

# IO ---


def _load_json(jsonfile: Union[Path, str]) -> dict:
    return json.loads(Path(jsonfile).read_text())


# Time ---


def _ts_to_datetime(ts: int) -> datetime:
    return datetime.utcfromtimestamp(ts / 1000)


def _jsontime_to_datetime(time_str: str) -> datetime:
    return datetime.strptime(time_str, "%d-%m-%y_%H-%M-%S")


def _datetime_to_string(dt: datetime):
    return dt.strftime("  %Y-%m-%d %H:%M:%S")


def _time_formatters():
    return {UTCTIME: _datetime_to_string, LOCALTIME: _datetime_to_string}


def _utc_to_local(dt: datetime):
    return dt.replace(tzinfo=tz.gettz("UTC")).astimezone(tz.gettz("Europe/Paris"))


def _local_to_utc(dt: datetime):
    return dt.replace(tzinfo=tz.gettz("Europe/Paris")).astimezone(tz.gettz("UTC"))


# Script Mode ------------------------------------------------------------------


def _get_args():
    """Parse Commandline Arguments."""
    # argparse is a bit weird: you need to create the normal parser
    # AND a parent parser if you want to use the same arg in both subparsers.
    # Using the main parser also as parent will result in `function` being
    # always `None`.
    parser = argparse.ArgumentParser(description="KickGroups Functions")
    parent_parser = argparse.ArgumentParser()
    parent_parser.add_argument(
        "--root",
        type=str,
        required=False,
        default=KICKGROUPS_ROOT,
        dest="root",
        help="KickGroups Root-Directory",
    )
    subparsers = parser.add_subparsers(
        title="Functionality",
        dest="function",
    )
    # ----- Full KickGroup Parser ----- #
    parser_kickgroups = subparsers.add_parser(
        "kickgroups",
        parents=[parent_parser],
        add_help=False,
        description="KickGroups",
        help="List all KickGroups",
    )
    parser_kickgroups.add_argument(
        "--sort",
        type=str,
        dest="sort",
        help="Sort KickGroups",
        choices=[TIMESTAMP, KICKGROUP],
        default=TIMESTAMP,
    )
    # ---- KickGroup Info Parser ---- #
    parser_info = subparsers.add_parser(
        "kickgroup_info",
        parents=[parent_parser],
        add_help=False,
        description="KickGroup Info",
        help="Show the info of a given KickGroup",
    )
    parser_info.add_argument(
        "name",
        type=str,
        help="KickGroup name",
    )
    return parser.parse_args()


if __name__ == "__main__":
    options = _get_args()
    if options.function == "kickgroups":
        list_available_kickgroups(by=options.sort, root=options.root)

    if options.function == "kickgroup_info":
        kickgroup_info(kick_group=options.name, root=options.root)
